#  -*- coding: utf-8 -*-
# Atari constants from the atari800 project, GPL licensed.  Transformed using
# the find regex: \{(\".+\"),\s*(0x[0-9a-fA-F]+).*$ and replace regex: \2: \1,


class EmptyMemoryMap(object):
    name = "No Memory Map"

    rmemmap = {}
    wmemmap = {}

    @classmethod
    def get_name(cls, addr, write=False):
        if write:
            if addr in cls.wmemmap:
                return cls.wmemmap[addr]
        if addr in cls.rmemmap:
            return cls.rmemmap[addr]
        return ""

    def __contains__(self, addr):
        return addr in self.rmemmap or addr in self.wmemmap


class Atari800MemoryMap(EmptyMemoryMap):
    name = "Atari 400/800"

    rmemmap = {
        0x0000: "LNFLG",
        0x0001: "NGFLAG",
        0x0002: "CASINI",
        0x0003: "CASINI+1",
        0x0004: "RAMLO",
        0x0005: "RAMLO+1",
        0x0006: "TRAMSZ",
        0x0007: "CMCMD",
        0x0008: "WARMST",
        0x0009: "BOOT?",
        0x000a: "DOSVEC",
        0x000b: "DOSVEC+1",
        0x000c: "DOSINI",
        0x000d: "DOSINI+1",
        0x000e: "APPMHI",
        0x000f: "APPMHI+1",
        0x0010: "POKMSK",
        0x0011: "BRKKEY",
        0x0012: "RTCLOK",
        0x0013: "RTCLOK+1",
        0x0014: "RTCLOK+2",
        0x0015: "BUFADR",
        0x0016: "BUFADR+1",
        0x0017: "ICCOMT",
        0x0018: "DSKFMS",
        0x0019: "DSKFMS+1",
        0x001a: "DSKUTL",
        0x001b: "DSKUTL+1",
        0x001c: "ABUFPT",
        0x001d: "ABUFPT+1",
        0x001e: "ABUFPT+2",
        0x001f: "ABUFPT+3",
        0x0020: "ICHIDZ",
        0x0021: "ICDNOZ",
        0x0022: "ICCOMZ",
        0x0023: "ICSTAZ",
        0x0024: "ICBALZ",
        0x0025: "ICBAHZ",
        0x0026: "ICPTLZ",
        0x0027: "ICPTHZ",
        0x0028: "ICBLLZ",
        0x0029: "ICBLHZ",
        0x002a: "ICAX1Z",
        0x002b: "ICAX2Z",
        0x002c: "ICSPRZ",
        0x002d: "ICSPRZ+1",
        0x002e: "ICIDNO",
        0x002f: "CIOCHR",
        0x0030: "STATUS",
        0x0031: "CHKSUM",
        0x0032: "BUFRLO",
        0x0033: "BUFRHI",
        0x0034: "BFENLO",
        0x0035: "BFENHI",
        0x0036: "LTEMP",
        0x0037: "LTEMP+1",
        0x0038: "BUFRFL",
        0x0039: "RECVDN",
        0x003a: "XMTDON",
        0x003b: "CHKSNT",
        0x003c: "NOCKSM",
        0x003d: "BPTR",
        0x003e: "FTYPE",
        0x003f: "FEOF",
        0x0040: "FREQ",
        0x0041: "SOUNDR",
        0x0042: "CRITIC",
        0x0043: "FMSZPG",
        0x0044: "FMSZPG+1",
        0x0045: "FMSZPG+2",
        0x0046: "FMSZPG+3",
        0x0047: "FMSZPG+4",
        0x0048: "FMSZPG+5",
        0x0049: "FMSZPG+6",
        0x004a: "ZCHAIN",
        0x004b: "ZCHAIN+1",
        0x004c: "DSTAT",
        0x004d: "ATRACT",
        0x004e: "DRKMSK",
        0x004f: "COLRSH",
        0x0050: "TMPCHR",
        0x0051: "HOLD1",
        0x0052: "LMARGN",
        0x0053: "RMARGN",
        0x0054: "ROWCRS",
        0x0055: "COLCRS",
        0x0056: "COLCRS+1",
        0x0057: "DINDEX",
        0x0058: "SAVMSC",
        0x0059: "SAVMSC+1",
        0x005a: "OLDROW",
        0x005b: "OLDCOL",
        0x005c: "OLDCOL+1",
        0x005d: "OLDCHR",
        0x005e: "OLDADR",
        0x005f: "OLDADR+1",
        0x0060: "FKDEF",
        0x0061: "FKDEF+1",
        0x0062: "PALNTS",
        0x0063: "LOGCOL",
        0x0064: "ADRESS",
        0x0065: "ADRESS+1",
        0x0066: "TOADR",
        0x0067: "TOADR+1",
        0x0068: "SAVADR",
        0x0069: "SAVADR+1",
        0x006a: "RAMTOP",
        0x006b: "BUFCNT",
        0x006c: "BUFSTR",
        0x006d: "BUFSTR+1",
        0x006e: "BITMSK",
        0x006f: "SHFAMT",
        0x0070: "ROWAC",
        0x0071: "ROWAC+1",
        0x0072: "COLAC",
        0x0073: "COLAC+1",
        0x0074: "ENDPT",
        0x0075: "ENDPT+1",
        0x0076: "DELTAR",
        0x0077: "DELTAC",
        0x0078: "DELTAC+1",
        0x0079: "KEYDEF",
        0x007a: "KEYDEF+1",
        0x007b: "SWPFLG",
        0x007c: "HOLDCH",
        0x007d: "INSDAT",
        0x007e: "COUNTR",
        0x007f: "COUNTR+1",
        0x0080: "LOMEM",
        0x0081: "LOMEM+1",
        0x0082: "VNTP",
        0x0083: "VNTP+1",
        0x0084: "VNTD",
        0x0085: "VNTD+1",
        0x0086: "VVTP",
        0x0087: "VVTP+1",
        0x0088: "STMTAB",
        0x0089: "STMTAB+1",
        0x008a: "STMCUR",
        0x008b: "STMCUR+1",
        0x008c: "STARP",
        0x008d: "STARP+1",
        0x008e: "RUNSTK",
        0x008f: "RUNSTK+1",
        0x0090: "MEMTOP",
        0x0091: "MEMTOP+1",
        0x0092: "MEOLFLG",
        0x0094: "COX",
        0x0095: "POKADR",
        0x0096: "POKADR+1",
        0x0097: "SVESA",
        0x0098: "SVESA+1",
        0x0099: "MVFA",
        0x009a: "MVFA+1",
        0x009b: "MVTA",
        0x009c: "MVTA+1",
        0x009d: "CPC",
        0x009e: "CPC+1",
        0x009f: "LLNGTH",
        0x00a0: "TSLNUM",
        0x00a1: "TSLNUM+1",
        0x00a2: "MVLNG",
        0x00a3: "MVLNG+1",
        0x00a4: "ECSIZE",
        0x00a5: "ECSIZE+1",
        0x00a6: "DIRFLG",
        0x00a7: "STMLBD",
        0x00a8: "STINDEX",
        0x00a9: "OPSTKX",
        0x00aa: "ARSTKX",
        0x00ab: "EXSVOP",
        0x00ac: "EXSVPR",
        0x00ad: "LELNUM",
        0x00ae: "LELNUM+1",
        0x00af: "STENUM",
        0x00b0: "COMCNT",
        0x00b1: "ADFLAG",
        0x00b2: "SVDISP",
        0x00b3: "ONLOOP",
        0x00b4: "ENTDTD",
        0x00b5: "LISTDTD",
        0x00b6: "DATAD",
        0x00b7: "DATALN",
        0x00b8: "DATALN+1",
        0x00b9: "ERRNUM",
        0x00ba: "STOPLN",
        0x00bb: "STOPLN+1",
        0x00bc: "TRAPLN",
        0x00bd: "TRAPLN+1",
        0x00be: "SAVCUR",
        0x00bf: "SAVCUR+1",
        0x00c0: "IOCMD",
        0x00c1: "IODVC",
        0x00c2: "PROMPT",
        0x00c3: "ERRSAV",
        0x00c4: "TEMPA",
        0x00c5: "TEMPA+1",
        0x00c6: "ZTEMP2",
        0x00c7: "ZTEMP2+1",
        0x00c8: "COLOR",
        0x00c9: "PTABW",
        0x00ca: "LOADFLG",
        0x00d2: "VTYPE",
        0x00d3: "VNUM",
        0x00d4: "FR0",
        0x00d5: "FR0+1",
        0x00d6: "FR0+2",
        0x00d7: "FR0+3",
        0x00d8: "FR0+4",
        0x00d9: "FR0+5",
        0x00da: "FRE",
        0x00db: "FRE+1",
        0x00dc: "FRE+2",
        0x00dd: "FRE+3",
        0x00de: "FRE+4",
        0x00df: "FRE+5",
        0x00e0: "FR1",
        0x00e1: "FR1+1",
        0x00e2: "FR1+2",
        0x00e3: "FR1+3",
        0x00e4: "FR1+4",
        0x00e5: "FR1+5",
        0x00e6: "FR2",
        0x00e7: "FR2+1",
        0x00e8: "FR2+2",
        0x00e9: "FR2+3",
        0x00ea: "FR2+4",
        0x00eb: "FR2+5",
        0x00ec: "FRX",
        0x00ed: "EEXP",
        0x00ee: "NSIGN",
        0x00ef: "ESIGN",
        0x00f0: "FCHRFLG",
        0x00f1: "DIGRT",
        0x00f2: "CIX",
        0x00f3: "INBUFF",
        0x00f4: "INBUFF+1",
        0x00f5: "ZTEMP1",
        0x00f6: "ZTEMP1+1",
        0x00f7: "ZTEMP4",
        0x00f8: "ZTEMP4+1",
        0x00f9: "ZTEMP3",
        0x00fa: "ZTEMP3+1",
        0x00fb: "RADFLG",
        0x00fc: "FLPTR",
        0x00fd: "FLPTR+1",
        0x00fe: "FPTR2",
        0x00ff: "FPTR2+1",
        0x0200: "VDSLST",
        0x0201: "VDSLST+1",
        0x0202: "VPRCED",
        0x0203: "VPRCED+1",
        0x0204: "VINTER",
        0x0205: "VINTER+1",
        0x0206: "VBREAK",
        0x0207: "VBREAK+1",
        0x0208: "VKEYBD",
        0x0209: "VKEYBD+1",
        0x020a: "VSERIN",
        0x020b: "VSERIN+1",
        0x020c: "VSEROR",
        0x020d: "VSEROR+1",
        0x020e: "VSEROC",
        0x020f: "VSEROC+1",
        0x0210: "VTIMR1",
        0x0211: "VTIMR1+1",
        0x0212: "VTIMR2",
        0x0213: "VTIMR2+1",
        0x0214: "VTIMR4",
        0x0215: "VTIMR4+1",
        0x0216: "VIMIRQ",
        0x0217: "VIMIRQ+1",
        0x0218: "CDTMV1",
        0x0219: "CDTMV1+1",
        0x021a: "CDTMV2",
        0x021b: "CDTMV2+1",
        0x021c: "CDTMV3",
        0x021d: "CDTMV3+1",
        0x021e: "CDTMV4",
        0x021f: "CDTMV4+1",
        0x0220: "CDTMV5",
        0x0221: "CDTMV5+1",
        0x0222: "VVBLKI",
        0x0223: "VVBLKI+1",
        0x0224: "VVBLKD",
        0x0225: "VVBLKD+1",
        0x0226: "CDTMA1",
        0x0227: "CDTMA1+1",
        0x0228: "CDTMA2",
        0x0229: "CDTMA2+1",
        0x022a: "CDTMF3",
        0x022b: "SRTIMR",
        0x022c: "CDTMF4",
        0x022d: "INTEMP",
        0x022e: "CDTMF5",
        0x022f: "SDMCTL",
        0x0230: "SDLSTL",
        0x0231: "SDLSTH",
        0x0232: "SSKCTL",
        0x0233: "LCOUNT",
        0x0234: "LPENH",
        0x0235: "LPENV",
        0x0236: "BRKKY",
        0x0237: "BRKKY+1",
        0x0238: "VPIRQ",
        0x0239: "VPIRQ+1",
        0x023a: "CDEVIC",
        0x023b: "CCOMND",
        0x023c: "CAUX1",
        0x023d: "CAUX2",
        0x023e: "TEMP",
        0x023f: "ERRFLG",
        0x0240: "DFLAGS",
        0x0241: "DBSECT",
        0x0242: "BOOTAD",
        0x0243: "BOOTAD+1",
        0x0244: "COLDST",
        0x0245: "RECLEN",
        0x0246: "DSKTIM",
        0x0247: "PDVMSK",
        0x0248: "SHPDVS",
        0x0249: "PDIMSK",
        0x024a: "RELADR",
        0x024b: "RELADR+1",
        0x024c: "PPTMPA",
        0x024d: "PPTMPX",
        0x026b: "CHSALT",
        0x026c: "VSFLAG",
        0x026d: "KEYDIS",
        0x026e: "FINE",
        0x026f: "GPRIOR",
        0x0270: "PADDL0",
        0x0271: "PADDL1",
        0x0272: "PADDL2",
        0x0273: "PADDL3",
        0x0274: "PADDL4",
        0x0275: "PADDL5",
        0x0276: "PADDL6",
        0x0277: "PADDL7",
        0x0278: "STICK0",
        0x0279: "STICK1",
        0x027a: "STICK2",
        0x027b: "STICK3",
        0x027c: "PTRIG0",
        0x027d: "PTRIG1",
        0x027e: "PTRIG2",
        0x027f: "PTRIG3",
        0x0280: "PTRIG4",
        0x0281: "PTRIG5",
        0x0282: "PTRIG6",
        0x0283: "PTRIG7",
        0x0284: "STRIG0",
        0x0285: "STRIG1",
        0x0286: "STRIG2",
        0x0287: "STRIG3",
        0x0288: "HIBYTE",
        0x0289: "WMODE",
        0x028a: "BLIM",
        0x028b: "IMASK",
        0x028c: "JVECK",
        0x028d: "JVECK+1",
        0x028e: "NEWADR",
        0x028f: "NEWADR+1",
        0x0290: "TXTROW",
        0x0291: "TXTCOL",
        0x0292: "TXTCOL+1",
        0x0293: "TINDEX",
        0x0294: "TXTMSC",
        0x0295: "TXTMSC+1",
        0x0296: "TXTOLD",
        0x0297: "TXTOLD+1",
        0x0298: "TXTOLD+2",
        0x0299: "TXTOLD+3",
        0x029a: "TXTOLD+4",
        0x029b: "TXTOLD+5",
        0x029c: "CRETRY",
        0x029d: "HOLD3",
        0x029e: "SUBTMP",
        0x029f: "HOLD2",
        0x02a0: "DMASK",
        0x02a1: "TMPLBT",
        0x02a2: "ESCFLG",
        0x02a3: "TABMAP",
        0x02a4: "TABMAP+1",
        0x02a5: "TABMAP+2",
        0x02a6: "TABMAP+3",
        0x02a7: "TABMAP+4",
        0x02a8: "TABMAP+5",
        0x02a9: "TABMAP+6",
        0x02aa: "TABMAP+7",
        0x02ab: "TABMAP+8",
        0x02ac: "TABMAP+9",
        0x02ad: "TABMAP+A",
        0x02ae: "TABMAP+B",
        0x02af: "TABMAP+C",
        0x02b0: "TABMAP+D",
        0x02b1: "TABMAP+E",
        0x02b2: "LOGMAP",
        0x02b3: "LOGMAP+1",
        0x02b4: "LOGMAP+2",
        0x02b5: "LOGMAP+3",
        0x02b6: "INVFLG",
        0x02b7: "FILFLG",
        0x02b8: "TMPROW",
        0x02b9: "TMPCOL",
        0x02ba: "TMPCOL+1",
        0x02bb: "SCRFLG",
        0x02bc: "HOLD4",
        0x02bd: "DRETRY",
        0x02be: "SHFLOK",
        0x02bf: "BOTSCR",
        0x02c0: "PCOLR0",
        0x02c1: "PCOLR1",
        0x02c2: "PCOLR2",
        0x02c3: "PCOLR3",
        0x02c4: "COLOR0",
        0x02c5: "COLOR1",
        0x02c6: "COLOR2",
        0x02c7: "COLOR3",
        0x02c8: "COLOR4",
        0x02c9: "RUNADR",
        0x02ca: "RUNADR+1",
        0x02cb: "HIUSED",
        0x02cc: "HIUSED+1",
        0x02cd: "ZHIUSE",
        0x02ce: "ZHIUSE+1",
        0x02cf: "GBYTEA",
        0x02d0: "GBYTEA+1",
        0x02d1: "LOADAD",
        0x02d2: "LOADAD+1",
        0x02d3: "ZLOADA",
        0x02d4: "ZLOADA+1",
        0x02d5: "DSCTLN",
        0x02d6: "DSCTLN+1",
        0x02d7: "ACMISR",
        0x02d8: "ACMISR+1",
        0x02d9: "KRPDEL",
        0x02da: "KEYREP",
        0x02db: "NOCLIK",
        0x02dc: "HELPFG",
        0x02dd: "DMASAV",
        0x02de: "PBPNT",
        0x02df: "PBUFSZ",
        0x02e0: "RUNAD",
        0x02e1: "RUNAD+1",
        0x02e2: "INITAD",
        0x02e3: "INITAD+1",
        0x02e4: "RAMSIZ",
        0x02e5: "MEMTOP",
        0x02e6: "MEMTOP+1",
        0x02e7: "MEMLO",
        0x02e8: "MEMLO+1",
        0x02e9: "HNDLOD",
        0x02ea: "DVSTAT",
        0x02eb: "DVSTAT+1",
        0x02ec: "DVSTAT+2",
        0x02ed: "DVSTAT+3",
        0x02ee: "CBAUDL",
        0x02ef: "CBAUDH",
        0x02f0: "CRSINH",
        0x02f1: "KEYDEL",
        0x02f2: "CH1",
        0x02f3: "CHACT",
        0x02f4: "CHBAS",
        0x02f5: "NEWROW",
        0x02f6: "NEWCOL",
        0x02f7: "NEWCOL+1",
        0x02f8: "ROWINC",
        0x02f9: "COLINC",
        0x02fa: "CHAR",
        0x02fb: "ATACHR",
        0x02fc: "CH",
        0x02fd: "FILDAT",
        0x02fe: "DSPFLG",
        0x02ff: "SSFLAG",
        0x0300: "DDEVIC",
        0x0301: "DUNIT",
        0x0302: "DCOMND",
        0x0303: "DSTATS",
        0x0304: "DBUFLO",
        0x0305: "DBUFHI",
        0x0306: "DTIMLO",
        0x0307: "DUNUSE",
        0x0308: "DBYTLO",
        0x0309: "DBYTHI",
        0x030a: "DAUX1",
        0x030b: "DAUX2",
        0x030c: "TIMER1",
        0x030d: "TIMER1+1",
        0x030e: "ADDCOR",
        0x030f: "CASFLG",
        0x0310: "TIMER2",
        0x0311: "TIMER2+1",
        0x0312: "TEMP1",
        0x0313: "TEMP2",
        0x0314: "PTIMOT",
        0x0315: "TEMP3",
        0x0316: "SAVIO",
        0x0317: "TIMFLG",
        0x0318: "STACKP",
        0x0319: "TSTAT",
        0x031a: "HATABS",
        0x033d: "PUPBT1",
        0x033e: "PUPBT2",
        0x033f: "PUPBT3",
        0x0340: "B0-ICHID",
        0x0341: "B0-ICDNO",
        0x0342: "B0-ICCOM",
        0x0343: "B0-ICSTA",
        0x0344: "B0-ICBAL",
        0x0345: "B0-ICBAH",
        0x0346: "B0-ICPTL",
        0x0347: "B0-ICPTH",
        0x0348: "B0-ICBLL",
        0x0349: "B0-ICBLH",
        0x034a: "B0-ICAX1",
        0x034b: "B0-ICAX2",
        0x034c: "B0-ICAX3",
        0x034d: "B0-ICAX4",
        0x034e: "B0-ICAX5",
        0x034f: "B0-ICAX6",
        0x0350: "B1-ICHID",
        0x0351: "B1-ICDNO",
        0x0352: "B1-ICCOM",
        0x0353: "B1-ICSTA",
        0x0354: "B1-ICBAL",
        0x0355: "B1-ICBAH",
        0x0356: "B1-ICPTL",
        0x0357: "B1-ICPTH",
        0x0358: "B1-ICBLL",
        0x0359: "B1-ICBLH",
        0x035a: "B1-ICAX1",
        0x035b: "B1-ICAX2",
        0x035c: "B1-ICAX3",
        0x035d: "B1-ICAX4",
        0x035e: "B1-ICAX5",
        0x035f: "B1-ICAX6",
        0x0360: "B2-ICHID",
        0x0361: "B2-ICDNO",
        0x0362: "B2-ICCOM",
        0x0363: "B2-ICSTA",
        0x0364: "B2-ICBAL",
        0x0365: "B2-ICBAH",
        0x0366: "B2-ICPTL",
        0x0367: "B2-ICPTH",
        0x0368: "B2-ICBLL",
        0x0369: "B2-ICBLH",
        0x036a: "B2-ICAX1",
        0x036b: "B2-ICAX2",
        0x036c: "B2-ICAX3",
        0x036d: "B2-ICAX4",
        0x036e: "B2-ICAX5",
        0x036f: "B2-ICAX6",
        0x0370: "B3-ICHID",
        0x0371: "B3-ICDNO",
        0x0372: "B3-ICCOM",
        0x0373: "B3-ICSTA",
        0x0374: "B3-ICBAL",
        0x0375: "B3-ICBAH",
        0x0376: "B3-ICPTL",
        0x0377: "B3-ICPTH",
        0x0378: "B3-ICBLL",
        0x0379: "B3-ICBLH",
        0x037a: "B3-ICAX1",
        0x037b: "B3-ICAX2",
        0x037c: "B3-ICAX3",
        0x037d: "B3-ICAX4",
        0x037e: "B3-ICAX5",
        0x037f: "B3-ICAX6",
        0x0380: "B4-ICHID",
        0x0381: "B4-ICDNO",
        0x0382: "B4-ICCOM",
        0x0383: "B4-ICSTA",
        0x0384: "B4-ICBAL",
        0x0385: "B4-ICBAH",
        0x0386: "B4-ICPTL",
        0x0387: "B4-ICPTH",
        0x0388: "B4-ICBLL",
        0x0389: "B4-ICBLH",
        0x038a: "B4-ICAX1",
        0x038b: "B4-ICAX2",
        0x038c: "B4-ICAX3",
        0x038d: "B4-ICAX4",
        0x038e: "B4-ICAX5",
        0x038f: "B4-ICAX6",
        0x0390: "B5-ICHID",
        0x0391: "B5-ICDNO",
        0x0392: "B5-ICCOM",
        0x0393: "B5-ICSTA",
        0x0394: "B5-ICBAL",
        0x0395: "B5-ICBAH",
        0x0396: "B5-ICPTL",
        0x0397: "B5-ICPTH",
        0x0398: "B5-ICBLL",
        0x0399: "B5-ICBLH",
        0x039a: "B5-ICAX1",
        0x039b: "B5-ICAX2",
        0x039c: "B5-ICAX3",
        0x039d: "B5-ICAX4",
        0x039e: "B5-ICAX5",
        0x039f: "B5-ICAX6",
        0x03a0: "B6-ICHID",
        0x03a1: "B6-ICDNO",
        0x03a2: "B6-ICCOM",
        0x03a3: "B6-ICSTA",
        0x03a4: "B6-ICBAL",
        0x03a5: "B6-ICBAH",
        0x03a6: "B6-ICPTL",
        0x03a7: "B6-ICPTH",
        0x03a8: "B6-ICBLL",
        0x03a9: "B6-ICBLH",
        0x03aa: "B6-ICAX1",
        0x03ab: "B6-ICAX2",
        0x03ac: "B6-ICAX3",
        0x03ad: "B6-ICAX4",
        0x03ae: "B6-ICAX5",
        0x03af: "B6-ICAX6",
        0x03b0: "B7-ICHID",
        0x03b1: "B7-ICDNO",
        0x03b2: "B7-ICCOM",
        0x03b3: "B7-ICSTA",
        0x03b4: "B7-ICBAL",
        0x03b5: "B7-ICBAH",
        0x03b6: "B7-ICPTL",
        0x03b7: "B7-ICPTH",
        0x03b8: "B7-ICBLL",
        0x03b9: "B7-ICBLH",
        0x03ba: "B7-ICAX1",
        0x03bb: "B7-ICAX2",
        0x03bc: "B7-ICAX3",
        0x03bd: "B7-ICAX4",
        0x03be: "B7-ICAX5",
        0x03bf: "B7-ICAX6",
        0x03c0: "PRNBUF",
        0x03e8: "SUPERF",
        0x03e9: "CKEY",
        0x03ea: "CASSBT",
        0x03eb: "CARTCK",
        0x03ec: "DERRF",
        0x03ed: "ACMVAR",
        0x03f8: "BASICF",
        0x03f9: "MINTLK",
        0x03fa: "GINTLK",
        0x03fb: "CHLINK",
        0x03fc: "CHLINK+1",
        0x03fd: "CASBUF",
        0x9ffa: "R-CARTCS",
        0x9ffb: "R-CARTCS+1",
        0x9ffc: "R-CART",
        0x9ffd: "R-CARTFG",
        0x9ffe: "R-CARTAD",
        0x9fff: "R-CARTAD+1",
        0xbffa: "CARTCS",
        0xbffb: "CARTCS+1",
        0xbffc: "CART",
        0xbffd: "CARTFG",
        0xbffe: "CARTAD",
        0xbfff: "CARTAD+1",
        0xd000: "M0PF",
        0xd001: "M1PF",
        0xd002: "M2PF",
        0xd003: "M3PF",
        0xd004: "P0PF",
        0xd005: "P1PF",
        0xd006: "P2PF",
        0xd007: "P3PF",
        0xd008: "M0PL",
        0xd009: "M1PL",
        0xd00a: "M2PL",
        0xd00b: "M3PL",
        0xd00c: "P0PL",
        0xd00d: "P1PL",
        0xd00e: "P2PL",
        0xd00f: "P3PL",
        0xd010: "TRIG0",
        0xd011: "TRIG1",
        0xd012: "TRIG2",
        0xd013: "TRIG3",
        0xd014: "PAL",
        0xd015: "COLPM3",
        0xd016: "COLPF0",
        0xd017: "COLPF1",
        0xd018: "COLPF2",
        0xd019: "COLPF3",
        0xd01a: "COLBK",
        0xd01b: "PRIOR",
        0xd01c: "VDELAY",
        0xd01d: "GRACTL",
        0xd01e: "HITCLR",
        0xd01f: "CONSOL",
        0xd100: "PBI",
        0xd1ff: "PDVI",
        0xd1ff: "PDVS",
        0xd200: "POT0",
        0xd201: "POT1",
        0xd202: "POT2",
        0xd203: "POT3",
        0xd204: "POT4",
        0xd205: "POT5",
        0xd206: "POT6",
        0xd207: "POT7",
        0xd208: "ALLPOT",
        0xd209: "KBCODE",
        0xd20a: "RANDOM",
        0xd20b: "POTGO",
        0xd20d: "SERIN",
        0xd20e: "IRQST",
        0xd20f: "SKSTAT",
        0xd300: "PORTA",
        0xd301: "PORTB",
        0xd302: "PACTL",
        0xd303: "PBCTL",
        0xd400: "DMACTL",
        0xd401: "CHACTL",
        0xd402: "DLISTL",
        0xd403: "DLISTH",
        0xd404: "HSCROL",
        0xd405: "VSCROL",
        0xd407: "PMBASE",
        0xd409: "CHBASE",
        0xd40a: "WSYNC",
        0xd40b: "VCOUNT",
        0xd40c: "PENH",
        0xd40d: "PENV",
        0xd40e: "NMIEN",
        0xd40f: "NMIST",
        0xd40f: "NMIRES",
        0xd600: "PBIRAM",
        0xd800: "AFP",
        0xd803: "PDID1",
        0xd805: "PDIOV",
        0xd806: "PDIOV+1",
        0xd808: "PDIRQV",
        0xd809: "PDIRQV+1",
        0xd80b: "PDID2",
        0xd80d: "PDVV",
        0xd8e6: "FASC",
        0xd9aa: "IFP",
        0xd9d2: "FPI",
        0xda44: "ZFR0",
        0xda46: "ZF1",
        0xda60: "FSUB",
        0xda66: "FADD",
        0xdadb: "FMUL",
        0xdb28: "FDIV",
        0xdd40: "PLYEVL",
        0xdd89: "FLD0R",
        0xdd8d: "FLD0P",
        0xdd98: "FLD1R",
        0xdd9c: "FLD1P",
        0xdda7: "FST0R",
        0xddab: "FST0P",
        0xddb6: "FMOVE",
        0xddc0: "EXP",
        0xddcc: "EXP10",
        0xdecd: "LOG",
        0xded1: "LOG10",
        0xe400: "EDITRV",
        0xe410: "SCRENV",
        0xe420: "KEYBDV",
        0xe430: "PRINTV",
        0xe440: "CASETV",
        0xe450: "DINITV",
        0xe453: "DSKINV",
        0xe456: "CIOV",
        0xe459: "SIOV",
        0xe45c: "SETVBV",
        0xe45f: "SYSVBV",
        0xe462: "XITVBV",
        0xe465: "SIOINV",
        0xe468: "SENDEV",
        0xe46b: "INTINV",
        0xe46e: "CIOINV",
        0xe471: "BLKBDV",
        0xe474: "WARMSV",
        0xe477: "COLDSV",
        0xe47a: "RBLOKV",
        0xe47d: "CSOPIV",
        0xe480: "PUPDIV",
        0xe483: "SLFTSV",
        0xe486: "PHENTV",
        0xe489: "PHUNLV",
        0xe48c: "PHINIV",
        0xe48f: "GPDVV",
    }

    wmemmap = {
        0xd000: "HPOSP0",
        0xd001: "HPOSP1",
        0xd002: "HPOSP2",
        0xd003: "HPOSP3",
        0xd004: "HPOSM0",
        0xd005: "HPOSM1",
        0xd006: "HPOSM2",
        0xd007: "HPOSM3",
        0xd008: "SIZEP0",
        0xd009: "SIZEP1",
        0xd00a: "SIZEP2",
        0xd00b: "SIZEP3",
        0xd00c: "SIZEM",
        0xd00d: "GRAFP0",
        0xd00e: "GRAFP1",
        0xd00f: "GRAFP2",
        0xd010: "GRAFP3",
        0xd011: "GRAFM",
        0xd012: "COLPM0",
        0xd013: "COLPM1",
        0xd014: "COLPM2",
        0xd200: "AUDF1",
        0xd201: "AUDC1",
        0xd202: "AUDF2",
        0xd203: "AUDC2",
        0xd204: "AUDF3",
        0xd205: "AUDC3",
        0xd206: "AUDF4",
        0xd207: "AUDC4",
        0xd208: "AUDCTL",
        0xd209: "STIMER",
        0xd20a: "SKRES",
        0xd20d: "SEROUT",
        0xd20e: "IRQEN",
        0xd20f: "SKCTL",
        }


class Atari5200MemoryMap(Atari800MemoryMap):
    name = "Atari 5200"

    rmemmap = {
        0x0000: "POKMSK",
        0x0001: "RTCLOKH",
        0x0002: "RTCLOKL",
        0x0003: "CRITIC",
        0x0004: "ATRACT",
        0x0005: "SDLSTL",
        0x0006: "SDLSTH",
        0x0007: "SDMCTL",
        0x0008: "PCOLR0",
        0x0009: "PCOLR1",
        0x000a: "PCOLR2",
        0x000b: "PCOLR3",
        0x000c: "COLOR0",
        0x000d: "COLOR1",
        0x000e: "COLOR2",
        0x000f: "COLOR3",
        0x0010: "COLOR4",
        0x0011: "PADDL0",
        0x0012: "PADDL1",
        0x0013: "PADDL2",
        0x0014: "PADDL3",
        0x0015: "PADDL4",
        0x0016: "PADDL5",
        0x0017: "PADDL6",
        0x0018: "PADDL7",
        0x0200: "VIMIRQ",
        0x0201: "VIMIRQ+1",
        0x0202: "VVBLKI",
        0x0203: "VVBLKI+1",
        0x0204: "VVBLKD",
        0x0205: "VVBLKD+1",
        0x0206: "VDSLST",
        0x0207: "VDSLST+1",
        0x0208: "VKEYBD",
        0x0209: "VKEYBD+1",
        0x020a: "VKPD",
        0x020b: "VKPD+1",
        0x020c: "BRKKY",
        0x020d: "BRKKY+1",
        0x020e: "VBREAK",
        0x020f: "VBREAK+1",
        0x0210: "VSERIN",
        0x0211: "VSERIN+1",
        0x0212: "VSEROR",
        0x0213: "VSEROR+1",
        0x0214: "VSEROC",
        0x0215: "VSEROC+1",
        0x0216: "VTIMR1",
        0x0217: "VTIMR1+1",
        0x0218: "VTIMR2",
        0x0219: "VTIMR2+1",
        0x021a: "VTIMR4",
        0x021b: "VTIMR4+1",
        0xc000: "M0PF",
        0xc001: "M1PF",
        0xc002: "M2PF",
        0xc003: "M3PF",
        0xc004: "P0PF",
        0xc005: "P1PF",
        0xc006: "P2PF",
        0xc007: "P3PF",
        0xc008: "M0PL",
        0xc009: "M1PL",
        0xc00a: "M2PL",
        0xc00b: "M3PL",
        0xc00c: "P0PL",
        0xc00d: "P1PL",
        0xc00e: "P2PL",
        0xc00f: "P3PL",
        0xc010: "TRIG0",
        0xc011: "TRIG1",
        0xc012: "TRIG2",
        0xc013: "TRIG3",
        0xc014: "PAL",
        0xc015: "COLPM3",
        0xc016: "COLPF0",
        0xc017: "COLPF1",
        0xc018: "COLPF2",
        0xc019: "COLPF3",
        0xc01a: "COLBK",
        0xc01b: "PRIOR",
        0xc01c: "VDELAY",
        0xc01d: "GRACTL",
        0xc01e: "HITCLR",
        0xc01f: "CONSOL",
        0xd400: "DMACTL",
        0xd401: "CHACTL",
        0xd402: "DLISTL",
        0xd403: "DLISTH",
        0xd404: "HSCROL",
        0xd405: "VSCROL",
        0xd407: "PMBASE",
        0xd409: "CHBASE",
        0xd40a: "WSYNC",
        0xd40b: "VCOUNT",
        0xd40c: "PENH",
        0xd40d: "PENV",
        0xd40e: "NMIEN",
        0xd40f: "NMIST",
        0xd40f: "NMIRES",
        0xe800: "POT0",
        0xe801: "POT1",
        0xe802: "POT2",
        0xe803: "POT3",
        0xe804: "POT4",
        0xe805: "POT5",
        0xe806: "POT6",
        0xe807: "POT7",
        0xe808: "ALLPOT",
        0xe809: "KBCODE",
        0xe80a: "RANDOM",
        0xe80b: "POTGO",
        0xe80d: "SERIN",
        0xe80e: "IRQST",
        0xe80f: "SKSTAT",
        }

    wmemmap = {
        0xc000: "HPOSP0",
        0xc001: "HPOSP1",
        0xc002: "HPOSP2",
        0xc003: "HPOSP3",
        0xc004: "HPOSM0",
        0xc005: "HPOSM1",
        0xc006: "HPOSM2",
        0xc007: "HPOSM3",
        0xc008: "SIZEP0",
        0xc009: "SIZEP1",
        0xc00a: "SIZEP2",
        0xc00b: "SIZEP3",
        0xc00c: "SIZEM",
        0xc00d: "GRAFP0",
        0xc00e: "GRAFP1",
        0xc00f: "GRAFP2",
        0xc010: "GRAFP3",
        0xc011: "GRAFM",
        0xc012: "COLPM0",
        0xc013: "COLPM1",
        0xc014: "COLPM2",
        0xe800: "AUDF1",
        0xe801: "AUDC1",
        0xe802: "AUDF2",
        0xe803: "AUDC2",
        0xe804: "AUDF3",
        0xe805: "AUDC3",
        0xe806: "AUDF4",
        0xe807: "AUDC4",
        0xe808: "AUDCTL",
        0xe809: "STIMER",
        0xe80a: "SKRES",
        0xe80d: "SEROUT",
        0xe80e: "IRQEN",
        0xe80f: "SKCTL",
        }


class Apple2MemoryMap(EmptyMemoryMap):
    name = "Apple ]["

    rmemmap = {
 #Monitor Zero Page locations
        0x0020: "WNDLFT",       # Left column of the Scroll Window
        0x0021: "WNDWDTH",       # Width of the Scroll Window
        0x0022: "WNDTOP",       # Top line of the Scroll Window
        0x0023: "WNDBTM",       # Bottom line of Scroll Window
        0x0024: "CH",       # Displacement from WNDLFT
        0x0025: "CV",
        0x0026: "GBASL",
        0x0027: "GBASH",
        0x0028: "BASL",
        0x0029: "BASH",
        0x002A: "BAS2L",
        0x002B: "BAS2H",
        0x002C: "H2",
        0x002D: "V2",
        0x002F: "LASTIN",
        0x0030: "COLOR",
        0x0031: "MODE",
        0x0032: "INVFLG",
        0x0033: "PROMPT",
        0x0034: "YSAV",
        0x0035: "YSAV1",
        0x0036: "CSWL",
        0x0037: "CSWH",
        0x0038: "KSWL",
        0x0039: "KSWH",
        0x003A: "PCL",
        0x003B: "PCH",
        0x003C: "A1L",
        0x003D: "A1H",
        0x003E: "A2L",
        0x003F: "A2H",
        0x0040: "A3L",
        0x0041: "A3H",
        0x0042: "A4L",
        0x0043: "A4H",
        0x0044: "A5L",
        0x0045: "A5H",
        0x0046: "XREG",
        0x0047: "YREG",
        0x0048: "STATUS",
        0x0049: "SPNT",
        0x004E: "RNDL",
        0x004F: "RNDH",
        0x0050: "ACL",
        0x0051: "ACH",
        0x0052: "XTNDL",
        0x0053: "XTNDH",
        0x0054: "AUXL",
        0x0055: "AUXH",
# IO registers
        0xC000: "KEYBOARD",       # keyboard data (latched) (RD-only)
        0xC001: "SET80COL",
        0xC002: "CLRAUXRD",       # read from auxilliary 48K
        0xC003: "SETAUXRD",
        0xC004: "CLRAUXWR",       # write to auxilliary 48K
        0xC005: "SETAUXWR",
        0xC006: "CLRCXROM",       # use external slot ROM
        0xC007: "SETCXROM",
        0xC008: "CLRAUXZP",       # use auxilliary ZP, stack, & LC
        0xC009: "SETAUXZP",
        0xC00A: "CLRC3ROM",       # use external slot C3 ROM
        0xC00B: "SETC3ROM",
        0xC00C: "CLR80VID",       # use 80-column display mode
        0xC00D: "SET80VID",
        0xC00E: "CLRALTCH",       # use alternate character set ROM
        0xC00F: "SETALTCH",
        0xC010: "STROBE",       # strobe (unlatch) keyboard data
        0xC011: "RDLCBNK2",       # reading from LC bank $Dx 2
        0xC012: "RDLCRAM",       # reading from LC RAM
        0xC013: "RDRAMRD",       # reading from auxilliary 48K
        0xC014: "RDRAMWR",       # writing to auxilliary 48K
        0xC015: "RDCXROM",       # using external slot ROM
        0xC016: "RDAUXZP",       # using auxilliary ZP, stack, & LC
        0xC017: "RDC3ROM",       # using external slot C3 ROM
        0xC018: "RD80COL",       # using 80-column memory mapping
        0xC019: "RDVBLBAR",       # not VBL (VBL signal low)
        0xC01A: "RDTEXT",       # using text mode
        0xC01B: "RDMIXED",       # using mixed mode
        0xC01C: "RDPAGE2",       # using text/graphics page2
        0xC01D: "RDHIRES",       # using Hi-res graphics mode
        0xC01E: "RDALTCH",       # using alternate character set ROM
        0xC01F: "RD80VID",       # using 80-column display mode
        0xC030: "SPEAKER",       # toggle speaker diaphragm
        0xC050: "CLRTEXT",       # enable text-only mode
        0xC051: "SETTEXT",
        0xC052: "CLRMIXED",       # enable graphics/text mixed mode
        0xC053: "SETMIXED",
        0xC054: "TXTPAGE1",       # select page1/2 (or page1/1x)
        0xC055: "TXTPAGE2",
        0xC056: "CLRHIRES",       # enable Hi-res graphics
        0xC057: "SETHIRES",
        0xC058: "SETAN0",       # 4-bit annunciator inputs
        0xC059: "CLRAN0",
        0xC05A: "SETAN1",
        0xC05B: "CLRAN1",
        0xC05C: "SETAN2",
        0xC05D: "CLRAN2",
        0xC05E: "SETAN3",
        0xC05F: "CLRAN3",
        0xC060: "CASSETTE",			# Cassette input
        0xC061: "OPNAPPLE",       # open apple (command) key data
        0xC062: "CLSAPPLE",       # closed apple (option) key data
        0xC070: "PDLTRIG",       # trigger paddles
        0xC081: "ROMIN",       # RD ROM, WR-enable LC RAM
        0xC083: "LCBANK2",       # RD LC RAM bank2, WR-enable LC RAM
        0xC08B: "LCBANK1",       # RD LC RAM bank1, WR-enable LC RAM
        0xCFFF: "CLRC8ROM",       # switch out slot C8 ROM
#Keyboard-related ROM routines, from Monitors Peeled
        0xFD0C: "RDKEY",	# Set screen to blink at cursor saving original character in A-reg. from (BASL),Y
        0xFD18: "RDKEY",	# Jump Indirect (KSWL) to KEYIN
        0xFD1B: "KEYIN",	# Increment random number at RNDL,H while polling keyboard register.
        0xFD26: "KEYIN",	# Store A-reg to (BASL),Y (clear blink set by RDKEY routine).
        0xFD28: "KEYIN",	# Load A-reg from Keyboard register
        0xFD2F: "ESC",	# Call RDKEY for Escape key service
        0xFD32: "ESC",	# Call ESC1 with char in A-reg to do indicated function.
        0xFD35: "RDCHAR",	# Call RDKEY to get next char into A. Compare to $93. =, br to ESC to call for next char and do ESC
        0xFC2C: "ESC1",	# Using character in A-reg, br to FC2C 64556 -980 ESC1 routine for Escape key service.
        0xFD3D: "NOTCR",	#Echo keyboard input thru COUT to screen, from IN,X , with INVFLG temporarily set to $FF.
        0xFD4D: "NOTCR",	#Pickup char from IN,X; if $88 goto BCKSPC; if $98 goto CANCEL; if X-reg (input index) greater than $F7 fall into FD5C.; Otherwise to NOTCR1, bypass Bell
        0xFD5C: "NOTCR",	#Sound bell if X indicates 248+ input characters.
        0xFD5F: "NOTCR1",	#Increment X ; If X not zero goto NXTCHAR; If x=o fall into CANCEL
        0xFD62: "CANCEL",	#Load $DC (backslash) into A-reg to indicate cancelled input.
        0xFD64: "CANCEL",	#Call COUT to print A-reg then fall into GETLNZ
        0xFD67: "GETLNZ",	#Print Carriage Return thru COUT
        0xFD6A: "GETLN",	#Load PROMPT into A-reg
        0xFD6C: "GETLN",	#Call COUT to print A-reg
        0xFD6F: "GETLN",	#Load X-reg with $01 for passage thru backspace operation.
        0xFD71: "BCKSPC",	#If x=o goto GETLNZ to start over. else decrement X, fall into NXTCHAR
        0xFD75: "NXTCHAR",	#Call RDCHAR to get next character If character gotten is $95 (ctrlU cursor right arrow) pick up screen character from (BASL), Y to replace it.
        0xFD84: "ADDINP",	#store A-reg to input area at IN,X Compare to return.
#Cassette-related ROM routines, from Monitors Peeled
        0xFECD: "WRITE ",	#
        0xFEFD: "READ ",	#
        0xFCC9: "HEADR ",	#
        0xFCFA: "RD2BIT",	#
        0xFCFD: "RDBIT ",	#
        0xFCEC: "RDBYTE",	#
        0xFCD6: "WRBIT ",	#
        0xFEED: "WRBYTE",	#
#programming aids routines, from Monitors Peeled
        0xFDED: "COUT",	#Write byte in A to screen at CV,CH
        0xFD8E: "CROUT",	#Print Carriage Return thru COUT
        0xF948: "PRBLNK ",	#Print three blanks thru COUT
        0xF94A: "PRBL2",	#Print eX) blanks thru COUT
        0xF94C: "PRBL3",	#Print character in A followed by (X )-1 blanks
        0xFF3A: "BELL",	#Print BELL code thru COUT
        0xFF2D: "PRERR",	#Print "ERR" and BELL thru COUT
        0xFDE3: "PRHEX",	#Print low nibble of A as hex char
        0xFDDA: "PRBYTE",	#Print A-reg as 2 hex nibbles
        0xF940: "PRNTYX",	#Print hex of Y,X regs
        0xF941: "PRNTAX",	#Print hex of A,X regs
        0xF944: "PRNTX",	#Print hex of X reg
        0xFD96: "PRYX2",	#Print CR, then hex of Y,X regs, then minus sign or dash
        0xFD99: "PRHEX",	#Print hex of Y,X regs, then dash
        0xFD92: "PRA1",	#Print CR, hex of A1H,A1L, and dash
        0xFDA3: "XAM8",	#Print memory as hex with preceeding address from mmmm to mmm7 where mmmm is contents of A1L,H on entry.
        0xFDB3: "XAM",	#Print memory as hex from (A1L,H) to (A2L,H) inclusive.
        0xFF4A: "SAVE",	#Save A,X,Y,P,S regs at $45-49
        0xFAD7: "REGDSP",	#Display registers with names from  $45-49 as SAVEd, with preceeding carriage return.
        0xFADA: "RGDSP1",	#Display regs as above without CR
        0xFF3F: "RESTORE",	#Restore regs A,X,Y,P not S from $45
        0xFA43: "STEP",	#Execute one instruction at (PCL,H)
        0xFE2C: "MOVE",	#Move memory contents to (A4L,H) from (A1L,H) thru (A2L,H)
        0xFE36: "VFY",	#Compare memory contents (A4L,H) to (A1L,H) thru (A2L,H), differences are displayed.
        0xFCB4: "NXTA4",	#Increment A4L,H ($42,43) and
        0xFCBA: "NXTA1",	#Inc A1L,H (3C,D), set carry if A2L,H not less than A1L,H
# Output Subroutines, from Assembly Lines
        0xFDED: "COUT",	#Output a character
        0xFDF0: "COUT1",	#Output to screen
        0xFE80: "SETINV",	#Set Inverse mode
        0xFE84: "SETNORM",	#Set Normal Mode
        0xFD8E: "CROUT",	# Generate a <RETURN>
        0xFD8B: "CROUT1",	# <RETURN> with clear
        0xFDDA: "PRBYTE",	# Print a hexadecimal byte
        0xFDE3: "PRHEX",	# Print a hexadecimal digit
        0xF941: "PRNTAX",	# Print A and X in hexadecimal
        0xF948: "PRBLNK",	#Print 3 spaces
        0xF94A: "PRBL2",	#Print many spaces
        0xFF3A: "BELL",	#Output a 'bell' character
        0xFBDD: "BELL1",	# Bep the Apple's speaker
# Input Subroutines, from Assembly Lines
        0xFD0C: "RDKEY",	#Get an input character
        0xFD35: "RDCHAR",	#Get an input character or escape code
        0xFD1B: "KEYIN",	#Read the Apple's keyboard
        0xFD6A: "GETLN",	#Get an input line with prompt
        0xFD67: "GETLNZ",	#Get an input line
        0xFD6F: "GETLN1",	#Get an input line, no prompt
#Low-Res Graphics Subroutines, from Assembly Lines
        0xF864: "SETCOL",	#Set low-res graphics color
        0xF85F: "NEXTCOL",	#Increment colorby 3
        0xF800: "PLOT",		#Plot a block on the Low-Res Screen
        0xF819: "HLINE",	#Draw a horizontal line of blocks
        0xF828: "VLINE",	#Draw a vertical line of blocks
        0xF832: "CLRSCR",	#Clear the entire low-res screen
        0xF836: "CLRTOP",	#Clear the top ofthe low-res Screen
        0xF871: "SCRN",	#Read the low-res screen
# Hi-Res Graphics Subroutines, from Assembly Lines
        0xF3E2: "HGR ",	#Hi-res page 1
        0xF3D8: "HGR2",	#Hi-res page 2
        0xF3F2: "HCLR",	#Clear to black
        0xF3F6: "BKGND",	#Clear to color
        0xF6F0: "HCOLOR",	#Set color
        0xF411: "HPOSN",	#Position the cursor
        0xF457: "HPLOT",	#Plot at cursor
        0xF5CB: "HFIND",	#Return the cursor position
        0xF53A: "HLIN",	#Draw a line
        0xF730: "SHNUM",	#Load shape number
        0xF601: "DRAW",	#Draw a shape
        0xF65D: "XDRAW",	#Erase a shape (draw XOR)
# Floating Point Accumulator routines, from Assembly Lines
        0xEBAF: "ABS",	#Absolute value
        0xEC23: "INT",	#INT function
        0xEFAE: "RND",	#Random number
        0xEB82: "SIGN",	#Sign of FAC (in Accumulator)
        0xEB90: "SGN",	#Sign of FAC (in FAC)
        0xEE8D: "SQR",	#Square root
        0xEF09: "EXP",	#Exponentiation
        0xE941: "LOG",	#Logarithm base e
        0xEE97: "FPWRT",	#Raise ARG to the FAC power (base e)
        0xEBB2: "FCOMP",	#Compare FAC to memory
        0xEED0: "NEGOP",	#Multiply by -1
        0xE7A0: "FADDH",	#Add 0.5
        0xEA55: "DIV10",	#Divide by 10
        0xEA39: "MUL10",	#Multiply by 10
# Trig functions, from Assembly Lines
        0xEFEA: "COS",	#the cosine function of FAC.
        0xEFFA: "SIN",	#the sine function of FAC.
        0xEFF1: "TAN",	#the tangent function of FAC.
        0xF09E: "ATN",	#the arctangent of FAC.
        0xED34: "FOUT",	#Create a string at the start of the stack
# Other Subroutines, from Assembly Lines
        0xFCA8: "WAIT",	#Delay
        0xFB1E: "PREAD",	#Read a game controller
        0xFF2D: "PRERR",	#Print 'ERR'
        0xFF4A: "IOSAVE",	#Save all registers
        0xFF3F: "IOREST"	#Restore all registers
    }

    wmemmap = {
        0xC000: "CLR80COL"       # use 80-column memory mapping (WR-only)
    }


class KIM1MemoryMap(EmptyMemoryMap):
    name = "KIM-1"

    rmemmap = {
        0x00EF: "PCL"  ,     # Program Counter - Low Order Byte
        0x00F0: "PGH"  ,     # Program Counter - High Order Byte
        0x00F1: "P"    ,   # Status Register
        0x00F2: "SF"   ,    # Stack Pointer
        0x00F3: "A"    ,   # Accumulator
        0x00F4: "Y"    ,   # Y-Index Register
        0x00F5: "X"    ,   # X-Index Register
        0x1700: "PAD"  ,     # 6530-003 A Data Register
        0x1701: "PADD" ,      # 6530-003 A Data Direction Register
        0x1702: "PBD"  ,     # 6530-003 B Data Register
        0x1703: "PBDD" ,      # 6530-003 B Data Direction Register
        0x1704: "TIMER",       # 	 6530-003 Interval Timer
        0x170F: "TIMER2",       #
        0x17F5: "SAL"  ,     # Starting Address - Low Order Byte
        0x17F6: "SAH"  ,     # Starting Address - High Order Byte
        0x17F7: "EAL"  ,     # Ending Address - Low Order Byte
        0x17F8: "EAH"  ,     # Ending Address - High Order Byte
        0x17F9: "ID"   ,    # File Identification Number
        0x17FA: "NMIL" ,      # NMI Vector - Low Order Byte
        0x17FB: "NMIH" ,      # NMI Vector - High Order Byte
        0x17FC: "RSTL" ,      # RST Vector - Low Order Byte
        0x17FD: "RSTH" ,      # RST Vector - High Order Byte
        0x17FE: "IRQL" ,      # IRQ Vector - Low Order Byte
        0x17FF: "IRQH" ,      # IRQ Vector - High Order Byte
        0x1800: "DUMPT",       #	Start Address - Audio Tape Dump
        0x1873: "LOADT",       #Start Address - Audio Tape Load
        0x1C00: "NMI"  ,     #Start Address for NMI using KIM "Save Nachine" Routine (Load in 17FA & 17FB)
        0x17F7: "EAL"  ,     #Ending Address - Low Order Byte
        0x17F8: "EAH"       #Ending Address - High Order Byte
	}

    wmemmap = {
    }



import types
import uuid as stdlib_uuid

import numpy as np

from . import errors

import logging
log = logging.getLogger(__name__)
try:  # Expensive debugging
    _xd = _expensive_debugging
except NameError:
    _xd = False


def uuid():
    u = stdlib_uuid.uuid4()

    # Force it into a str type so it isn't serialized as something weird
    # through jsonpickle
    return str(u)


def to_numpy(value):
    if type(value) is np.ndarray:
        # force copy to make sure we aren't pointing to an immutable array
        return value.copy()
    elif type(value) is bytes:
        return np.copy(np.frombuffer(value, dtype=np.uint8))
    elif type(value) is list:
    	return np.asarray(value, dtype=np.uint8)
    raise TypeError("Can't convert to numpy data")


def to_numpy_list(value):
    if type(value) is np.ndarray:
        return value
    return np.asarray(value, dtype=np.uint32)


def text_to_int(text, default_base="hex"):
    """ Convert text to int, raising exeception on invalid input
    """
    if text.startswith("0x"):
        value = int(text[2:], 16)
    elif text.startswith("$"):
        value = int(text[1:], 16)
    elif text.startswith("#"):
        value = int(text[1:], 10)
    elif text.startswith("%"):
        value = int(text[1:], 2)
    else:
        if default_base == "dec":
            value = int(text)
        else:
            value = int(text, 16)
    return value


def bool_to_ranges(matches):
    w = np.where(matches == True)[0]
    # split into groups with consecutive numbers
    groups = np.split(w, np.where(np.diff(w) != 1)[0] + 1)
    ranges = []
    for group in groups:
        if np.alen(group) > 0:
            ranges.append((int(group[0]), int(group[-1]) + 1))
    return ranges


def collapse_values(src):
    """Given a list of integers, return a list of lists, where each entry
    contains a value and the start/end location of that value.

    For example, the list [0, 0, 0, 0, 255, 99, 99, 99] would return
    [
        [0, 0, 4],
        [255, 4, 5],
        [99, 5, 8],
    ]
    """
    d = to_numpy_list(src)
    changes = np.where(np.diff(d) > 0)[0]
    index = 0
    ranges = []
    for end in changes:
        end = end + 1
        ranges.append([int(d[index]), int(index), int(end)])
        index = end
    if index < len(d):
        ranges.append([int(d[index]), int(index), len(d)])
    return ranges


def restore_value_to_ranges(dest, ranges, value):
    """Restore a list given the description returned by `collapse_list`
    """
    for start, end in ranges:
        print(f"{hex(start)}:{hex(end)} = {value}")
        dest[start:end] = value


def restore_values(dest, ranges):
    """Restore a list given the description returned by `collapse_list`
    """
    for value, start, end in ranges:
        dest[start:end] = value


def collapse_to_ranges(src, compact=False):
    """Given a list of integers, return a list of lists that represent groups
    of monotonically increasing integers.

    For example, the list [0, 1, 2, 3, 4, 5, 7, 10, 11, 99, 500, 501, 502, 892]
    would return:

    [
        [0, 6],
        [7, 8],
        [10, 12],
        [99, 100],
        [500, 503],
        [892, 893],
    ]

    Note that ranges are returned in python slice notation; that is, the 2nd
    entry is one beyond the end value.

    if compact is True, one element ranges are stored as a single integer, so
    the above example would instead produce:

    [
        [0, 6],
        7,
        [10, 12],
        99,
        [500, 503],
        892,
    ]

    potentially saving space if there are a large number of entries with no
    neighbors.
    """
    groups = np.split(src, np.where(np.diff(src) != 1)[0] + 1)
    ranges = []
    for group in groups:
        if np.alen(group) > 0:
            start = int(group[0])
            end = int(group[-1]) + 1
            if compact and end == start + 1:
                ranges.append(start)
            else:
                ranges.append([start, end])
    return ranges


def restore_from_ranges(dest, ranges):
    index = 0
    for item in ranges:
        try:
            single = int(item)
        except TypeError:
            start, end = item
            count = end - start
            dest[index:index + count] = np.arange(start, end, dtype=np.uint32)
        else:
            count = 1
            dest[index:index + count] = single
        index += count



import sys

import wx

from sawx.preferences import SawxEditorPreferences
from sawx.ui.fonts import str_to_font, default_font

from ..editors.byte_editor_preferences import ByteEditorPreferences


class EmulatorPreferences(ByteEditorPreferences):
    def set_defaults(self):
        ByteEditorPreferences.set_defaults(self)
        self.default_emulator = "Atari 800"



# Standard library imports.
import sys
import os

# Major package imports.
import numpy as np
from atrip import style_bits
from atrip.machines.atari8bit.jumpman import parser as jp

# Local imports.
from sawx.utils.command import Overlay

from ..commands import jumpman as jc
from ..commands import ChangeByteCommand
from ..viewers.mouse_modes import NormalSelectMode

import logging
log = logging.getLogger(__name__)
selectlog = logging.getLogger("select")

class JumpmanSelectMode(NormalSelectMode):
    can_paste = False

    def __init__(self, *args, **kwargs):
        NormalSelectMode.__init__(self, *args, **kwargs)
        self.mouse_down = (0, 0)
        self.objects = []

    def cleanup(self):
        pass

    def resync_objects(self):
        pass

    @property
    def has_selection(self):
        return bool(self.objects)

    @property
    def can_copy(self):
        return self.can_paste and self.has_selection

    def all_objects_are_coins(self):
        if self.objects:
            state = True
            for obj in self.objects:
                if not obj.single:
                    state = False
                    break
        else:
            state = False
        return state

    def calc_playfield_override(self):
        """ Replace the entire bit image generation in JumpmanLevelView """
        return None

    def draw_extra_objects(self, lever_builder, screen, current_segment):
        return

    def draw_overlay(self, bitimage):
        return

    def get_harvest_offset(self):
        source = self.control.segment_viewer.segment
        if len(source) < 0x47:
            hx = hy = 0, 0
        else:
            hx = source[0x46]
            hy = source[0x47]
        return hx, hy

    def draw_harvest_grid(self, screen):
        hx, hy = self.get_harvest_offset()
        w = 160
        h = 88

        # Original (slow) algorithm to determine bad locations:
        #
        # def is_allergic(x, y, hx, hy):
        #     return (x + 0x30 + hx) & 0x1f < 7 or (2 * y + 0x20 + hy) & 0x1f < 5
        #
        # Note that in the originial 6502 code, the y coord is in player
        # coords, which is has twice the resolution of graphics 7. That's the
        # factor of two in the y part. Simplifying, the bad locations can be
        # defined in sets of 32 columns and 16 rows:
        #
        # x: 16 - hx, 16 - hx + 6 inclusive
        # y: 0 - hy/2, 0 - hy/2 + 2 inclusive
        hx = hx & 0x1f
        hy = hy & 0x1f

        # Fast numpy algorithmic way to generate a grid: create a function and
        # have it generate true/false values for each coordinate input pair.
        y, x = np.mgrid[:h,:w]
        allergic = np.logical_or((2 * y + 0x20 + hy) & 0x1f < 7, (x + 0x30 + hx) & 0x1f <= 7).astype(np.uint8) * style_bits.comment_bit_mask
        style = screen.style
        style |= allergic.flat


    def get_xy(self, evt):
        c = self.control
        y, x, _ = c.get_row_col_from_event(evt)
        if y < c.model.antic_lines:
            index, _ = c.table.get_index_range(y, x)
            pick = c.model.pick_buffer[index]
        else:
            pick = -1
        return x, y, pick

    def display_coords(self, evt, extra=None):
        c = self.control
        e = c.segment_viewer
        if e is not None:
            x, y, pick = self.get_xy(evt)
            msg = "x=%d (0x%x) y=%d (0x%x) pick=%d" % (x, x, y, y, pick)
            if extra:
                msg += " " + extra
            e.linked_base.editor.frame.status_message(msg)

    def process_left_down(self, evt):
        self.display_coords(evt)

    def process_left_up(self, evt):
        self.display_coords(evt)

    def process_left_dclick(self, evt):
        pass

    def process_mouse_motion_down(self, evt):
        self.display_coords(evt)

    def process_mouse_motion_up(self, evt):
        self.display_coords(evt)

    def get_picked(self, pick):
        return self.control.model.screen_state.get_picked(pick)

    def calc_popup_data(self, evt):
        cg = self.control
        row, col, _ = cg.get_row_col_from_event(evt)
        inside = True  # fixme
        style = 0
        obj = self.objects
        if len(obj) == 0:
            x, y, pick = self.get_xy(evt)
            if pick >= 0:
                p = self.get_picked(pick)
                if p.single:
                    obj = [p]
                else:
                    obj = None
            else:
                obj = None
        popup_data = {
            'index': None,
            'in_selection': style&0x80,
            'row': row,
            'col': col,
            'inside': inside,
            'jumpman_obj': obj,
            }
        return popup_data


class AnticDSelectMode(JumpmanSelectMode):
    icon = "cursor"
    menu_item_name = "Select"
    menu_item_tooltip = "Select regions"
    min_mouse_distance = 2
    can_paste = True

    def init_post_hook(self):
        self.pending_remove = None
        self.override_state = None

    def resync_objects(self):
        """ After a redraw when the level builder has been rebuilt from the hex
        data, the objects stored locally will no longer point to current
        jumpman objects. We need to find the current objects that match up to
        the stored objects.
        """
        selectlog.debug("before resyncing: %s" % self.objects)
        self.objects = self.control.model.level_builder.find_equivalent(self.objects)
        selectlog.debug("after resyncing: %s" % self.objects)

    def delete_objects(self):
        if self.objects:
            self.control.model.delete_objects(self.objects)
            self.objects = []
        self.control.Refresh()

    def calc_playfield_override(self):
        if not self.objects:
            self.override_state = None
            return

        model = self.control.model
        playfield = model.get_playfield_segment()  # use new, temporary playfield
        _, _, self.override_state = model.redraw_current(playfield, self.objects)
        log.debug("override_state: %s" % self.override_state)

        # Draw the harvest grid if a coin is selected
        for obj in self.objects:
            if obj.single:
                self.draw_harvest_grid(playfield)
                break
        return playfield

    def get_picked(self, pick):
        if self.override_state:
            state = self.override_state
        else:
            state = self.control.model.get_screen_state()
        if state is not None:
            return state.get_picked(pick)
        else:
            raise ValueError("Invalid state")

    def highlight_pick(self, evt):
        x, y, pick = self.get_xy(evt)
        self.mouse_down = x, y
        if pick >= 0:
            obj = self.get_picked(pick)
            selectlog.debug("picked object: %s" % obj)
            if obj in self.objects:
                if evt.ControlDown():
                    self.pending_remove = obj
                else:
                    self.pending_remove = True
            else:
                self.add_to_selection(obj, evt.ControlDown())
            self.check_tolerance = True
        else:
            # don't kill multiple selection if user clicks on empty space by
            # mistake
            if not evt.ControlDown():
                self.objects = []

    def add_to_selection(self, obj, append=False):
        obj.orig_x = int(obj.x)
        obj.orig_y = int(obj.y)
        if append:
            self.objects.append(obj)
        else:
            self.objects = [obj]

    def move_pick(self, evt):
        if self.objects:
            x, y, pick = self.get_xy(evt)
            dx = x - self.mouse_down[0]
            dy = y - self.mouse_down[1]
            if self.check_tolerance and abs(dx) + abs(dy) <  self.min_mouse_distance:
                return
            self.check_tolerance = False
            for obj in self.objects:
                selectlog.debug("moving %s, equiv %s" % (obj, self.control.model.level_builder.find_equivalent_object(obj)))
                obj.x = obj.orig_x + dx
                obj.x &= obj.valid_x_mask
                obj.y = obj.orig_y + dy
            self.pending_remove = None

    def process_left_down(self, evt):
        selectlog.debug("left down: %s" % evt)
        try:
            self.highlight_pick(evt)
        except ValueError:
            pass
        else:
            self.control.refresh_view()
            self.display_coords(evt)

    def process_mouse_motion_down(self, evt):
        if self.objects:
            self.move_pick(evt)
            self.control.refresh_view()
        self.display_coords(evt)

    def process_left_up(self, evt):
        if self.num_clicks == 2:
            return
        selectlog.debug("pending_remove: %s" % self.pending_remove)
        if self.pending_remove is True:
            self.objects = []
        elif self.pending_remove is not None:
            self.objects.remove(self.pending_remove)
        self.pending_remove = None
        if self.objects and not self.check_tolerance:
            self.control.segment_viewer.save_changes(keep_objects_selected=True)
        else:
            self.control.refresh_view()
        self.display_coords(evt)

    def process_mouse_motion_up(self, evt):
        self.display_coords(evt)

    def check_trigger_pick(self, evt):
        x, y, pick = self.get_xy(evt)
        if pick >= 0:
            obj = self.get_picked(pick)
            if obj.single:
                self.control.segment_viewer.set_trigger_view(obj)

    def process_left_dclick(self, evt):
        try:
            self.check_trigger_pick(evt)
        except ValueError:
            pass
        evt.Skip()

    def delete_key_pressed(self):
        self.delete_objects()

    def backspace_key_pressed(self):
        self.delete_objects()

    def get_popup_actions(self, evt):
        return self.get_trigger_popup_actions(evt)


class DrawMode(JumpmanSelectMode):
    icon = "cursor"
    menu_item_name = "Draw"
    menu_item_tooltip = "Draw stuff"
    drawing_object = jp.Girder
    can_paste = False

    def resync_objects(self):
        # objects here are only temporary, so no need to search
        pass

    def draw_extra_objects(self, level_builder, screen, current_segment):
        level_builder.draw_objects(screen, self.objects, current_segment)

    def create_objects(self, evt, start=False):
        c = self.control
        e = c.segment_viewer
        if e is None:
            return
        x, y, pick = self.get_xy(evt)
        if start:
            self.mouse_down = x, y
        dx = x - self.mouse_down[0]
        dy = y - self.mouse_down[1]
        obj = self.drawing_object
        if obj.vertical_only:
            sx = 0
            sy = obj.default_dy if dy > 0 else -obj.default_dy
            num = max((dy + sy - 1) // sy, 1)
        elif obj.single:
            sx = obj.default_dx
            sy = 0
            num = 1
        else:
            if abs(dx) >= abs(dy):
                sx = obj.default_dx if dx > 0 else -obj.default_dx
                num = max((abs(dx) + abs(sx) - 1) // abs(sx), 1)
                sy = dy // num
            else:
                sy = obj.default_dy if dy > 0 else -obj.default_dy
                num = max((abs(dy) + abs(sy) - 1) // abs(sy), 1)
                sx = dx // num
        screen_x = (self.mouse_down[0] - obj.default_dx // 2) & obj.valid_x_mask
        screen_y = self.mouse_down[1] - obj.default_dy // 2

        self.objects = []
        item = obj(-1, screen_x, screen_y, num, sx, sy)
        if not item.is_offscreen():
            self.objects.append(item)
        self.check_objects(x, y)

    def check_objects(self, x, y):
        pass

    def process_left_down(self, evt):
        self.create_objects(evt, True)
        self.control.refresh_view()
        self.display_coords(evt)

    def process_left_up(self, evt):
        if self.num_clicks == 2:
            return
        log.debug("saving objects: %s" % self.objects)
        self.control.model.save_objects(self.objects)
        self.objects = []
        self.control.refresh_view()
        self.display_coords(evt)

    def process_mouse_motion_down(self, evt):
        self.create_objects(evt)
        self.control.refresh_view()
        self.display_coords(evt)

    def process_mouse_motion_up(self, evt):
        self.create_objects(evt, True)
        self.control.refresh_view()
        self.display_coords(evt)


class DrawGirderMode(DrawMode):
    icon = "jumpman_girder"
    menu_item_name = "Draw Girder"
    menu_item_tooltip = "Draw girders"
    drawing_object = jp.Girder


class DrawDoubleGirderMode(DrawMode):
    icon = "jumpman_double_girder"
    menu_item_name = "Draw Double Girder"
    menu_item_tooltip = "Draw double-wide girders"
    drawing_object = jp.DoubleWideGirder


class DrawLadderMode(DrawMode):
    icon = "jumpman_ladder"
    menu_item_name = "Draw Ladder"
    menu_item_tooltip = "Draw ladders (vertical only)"
    drawing_object = jp.Ladder


class DrawUpRopeMode(DrawMode):
    icon = "jumpman_uprope"
    menu_item_name = "Draw Up Rope"
    menu_item_tooltip = "Draw up ropes (vertical only)"
    drawing_object = jp.UpRope


class DrawDownRopeMode(DrawMode):
    icon = "jumpman_downrope"
    menu_item_name = "Draw Down Rope"
    menu_item_tooltip = "Draw down ropes (vertical only)"
    drawing_object = jp.DownRope


class EraseGirderMode(DrawMode):
    icon = "jumpman_erase_girder"
    menu_item_name = "Erase Girder"
    menu_item_tooltip = "Erase girders"
    drawing_object = jp.EraseGirder
    editor_trait_for_enabled = 'focused_viewer.can_erase_objects'


class EraseLadderMode(DrawMode):
    icon = "jumpman_erase_ladder"
    menu_item_name = "Erase Ladder"
    menu_item_tooltip = "Erase ladders (vertical only)"
    drawing_object = jp.EraseLadder
    editor_trait_for_enabled = 'focused_viewer.can_erase_objects'


class EraseRopeMode(DrawMode):
    icon = "jumpman_erase_rope"
    menu_item_name = "Erase Rope"
    menu_item_tooltip = "Erase ropes (vertical only)"
    drawing_object = jp.EraseRope
    editor_trait_for_enabled = 'focused_viewer.can_erase_objects'


class DrawCoinMode(DrawMode):
    icon = "jumpman_coin"
    menu_item_name = "Draw Coins"
    menu_item_tooltip = "Draw coins (single only)"
    drawing_object = jp.Coin

    def __init__(self, *args, **kwargs):
        DrawMode.__init__(self, *args, **kwargs)
        self.is_bad_location = False
        self.batch = None

    def cleanup(self):
        v = self.control.segment_viewer
        v.current_level.playfield.style[:] &= (0xff ^ style_bits.comment_bit_mask)

    def get_caret(self):
        if self.is_bad_location:
            return wx.Cursor(wx.CURSOR_HAND)
        else:
            return wx.Cursor(wx.CURSOR_ARROW)

    def check_objects(self, x, y):
        hx, hy = self.get_harvest_offset()
        if self.objects:
            self.is_bad_location = self.objects[0].is_bad_location(hx, hy)
        else:
            self.is_bad_location = jp.is_bad_harvest_position(x, y, hx, hy)

    def draw_extra_objects(self, level_builder, screen, current_segment):
        if self.objects:
            obj = self.objects[0]
            obj.error = self.is_bad_location
            if not obj.error:
                hx, hy = self.get_harvest_offset()
                grid = obj.harvest_checksum(hx, hy)
                obj.error = grid in level_builder.harvest_offset_seen
        level_builder.draw_objects(screen, self.objects, current_segment)
        self.draw_harvest_grid(screen)

    def change_harvest_offset(self, evt, start=False):
        c = self.control
        e = c.segment_viewer
        if e is None:
            return
        x, y, pick = self.get_xy(evt)
        if start:
            self.batch = Overlay()
            hx, hy = self.get_harvest_offset()
            self.mouse_down = hx + x, hy + y
        else:
            dx = (self.mouse_down[0] - x) & 0x1f
            dy = ((self.mouse_down[1] - y) & 0xf) * 2
            self.display_coords(evt)
            cmd = jc.MoveHarvestGridCommand(e.segment, dx, dy)
            e.editor.process_command(cmd, self.batch)

    def process_left_down(self, evt):
        self.create_objects(evt, True)
        if self.is_bad_location:
            self.change_harvest_offset(evt, True)
            self.objects = []
        else:
            self.control.Refresh()
        self.display_coords(evt)

    def process_left_up(self, evt):
        if self.num_clicks == 2:
            return
        if self.batch is not None:
            c = self.control
            e = c.segment_viewer
            if e is None:
                return
            e.editor.end_batch()
            self.batch = None

            # Force updating of the hex view
            e.document.change_count += 1
            e.document.refresh_event(flags=True)
        else:
            DrawMode.process_left_up(self, evt)
            return
        self.display_coords(evt)

    def process_mouse_motion_down(self, evt):
        if self.batch is not None:
            self.change_harvest_offset(evt)
        else:
            self.create_objects(evt)
            self.control.refresh_view()
        self.display_coords(evt)

    def process_mouse_motion_up(self, evt):
        self.create_objects(evt, True)
        self.control.refresh_view()
        self.display_coords(evt)

    def get_popup_actions(self, evt):
        return self.get_trigger_popup_actions(evt)


class JumpmanRespawnMode(DrawMode):
    icon = "jumpman_respawn"
    menu_item_name = "Set Jumpman Start"
    menu_item_tooltip = "Set jumpman respawn position"
    drawing_object = jp.JumpmanRespawn

    def init_post_hook(self):
        x, y = self.get_respawn_point()
        self.current = self.drawing_object(-1, x, y, 1, 0, 0)
        self.objects = []

    def get_respawn_point(self):
        source = self.control.segment_viewer.segment
        x = source[0x39]
        y = source[0x3a]
        return x - 0x30, (y - 0x18) // 2

    def draw_extra_objects(self, level_builder, screen, current_segment):
        objects = [self.current]
        objects.extend(self.objects)
        level_builder.draw_objects(screen, objects, current_segment)

    def change_jumpman_respawn(self, evt):
        obj = self.objects[0]
        values = [obj.x + 0x30, (obj.y * 2) + 0x18]
        self.control.segment_viewer.linked_base.change_bytes(0x39, 0x3b, values, "Change Respawn Point")

    def process_left_up(self, evt):
        if self.num_clicks == 2:
            return
        self.change_jumpman_respawn(evt)
        self.init_post_hook()
        self.display_coords(evt)

    def process_mouse_motion_down(self, evt):
        self.process_mouse_motion_up(evt)



import os
import sys

import numpy as np

import wx

from atrip import style_bits

from sawx.utils.nputil import intscale
from sawx.ui import compactgrid as cg

from ..ui.segment_grid import SegmentGridControl, SegmentTable

from ..viewer import SegmentViewer
from . import bitmap as b
from ..arch import pixel_converters as px
from ..arch import colors

import logging
log = logging.getLogger(__name__)


class PixelLineRenderer(b.BitmapLineRenderer):
    """The difference between this and the normal bitmap renderer is this one
    is selectable on the pixel level, rather than the byte level. Since the
    data is stored in bytes, the column range calculations are different.

    The desired column range might start and/or end at a pixel in the middle of
    a byte, so the ranges and drawing rectangles must be expanded to byte
    boundaries.
    """
    def calc_cell_size_in_pixels(self, grid_control):
        w = grid_control.zoom_w
        h = grid_control.zoom_h
        return w, h

    def col_to_rect(self, line_num, col):
        cell = self._col_to_cell[col]
        x, y = self.cell_to_pixel(line_num, cell)
        w = self.pixel_widths[col]
        rect = wx.Rect(x, y, w, self.h)
        return rect

    def calc_column_range(self, parent, line_num, first_byte, last_byte):
        t = parent.table
        row_start = (line_num * t.bytes_per_row) - t.start_offset
        index = row_start + first_byte
        if index < 0:
            first_byte -= index
            index = 0
        last_index = row_start + last_byte
        if last_index > t.last_valid_index:
            last_index = t.last_valid_index
        if index >= last_index:
            raise IndexError("No items in this line are in the visible scrolled region")
        return first_byte, index, last_index

    # BaseLineRenderer interface

    def draw_line(self, grid_control, dc, line_num, first_byte, index, last_index):
        t = grid_control.table
        rect = self.col_to_rect(line_num, first_byte * t.pixels_per_byte)
        data = t.data[index:last_index]
        style = t.style[index:last_index]
        self.image_cache.draw_item(grid_control, dc, rect, data, style)

    def add_rectangular_selection(self, caret, style_per_pixel, t, first_row, last_row, first_byte, last_byte):
        # some part of selection is visible
        start_row = max(caret.anchor_start[0] - first_row, 0)
        first_col = first_byte * t.pixels_per_byte
        left_col = max(caret.anchor_start[1] - first_col, 0)
        last_col = last_byte * t.pixels_per_byte
        right_col = min(caret.anchor_end[1] - first_col + 1, last_col - first_col + 1)
        end_row = min(caret.anchor_end[0] - first_row + 1, last_row - first_row + 1)
        s2d = style_per_pixel.reshape((-1, (last_byte - first_byte) * style_per_pixel.shape[-1]))
        print("OETUSHNTOEHUSROEHU rectangular selection", start_row, left_col, end_row, right_col, style_per_pixel.shape, s2d.shape)
        s2d[start_row:end_row, left_col:right_col] |= style_bits.selected_bit_mask

    def draw_grid(self, grid_control, dc, first_row, visible_rows, first_cell, visible_cells):
        t = grid_control.table
        t.prepare_for_drawing(first_row, visible_rows, first_cell, visible_cells)
        last_cell = min(first_cell + visible_cells, self.num_cells)
        last_row = min(first_row + visible_rows, t.num_rows)
        log.debug("draw_grid: rows:%d,%d (vis %d, num %d) cols:%d,%d" % (first_row, last_row, visible_rows, t.num_rows, first_cell, last_cell))

        ul_rect = self.col_to_rect(first_row, first_cell)
        lr_rect = self.col_to_rect(last_row - 1, last_cell - 1)
        frame_rect = wx.Rect(ul_rect.x, ul_rect.y, lr_rect.x - ul_rect.x + lr_rect.width, lr_rect.y - ul_rect.y + lr_rect.height)
        dc.SetClippingRegion(frame_rect)
 
        pixels, style = t.current_rectangle
        pixels = pixels[:,first_cell:last_cell]
        style = style[:,first_cell:last_cell]
        array = px.calc_rgb_from_color_indexes(pixels, style, grid_control.color_list_rgb, grid_control.empty_color_rgb)
        width = array.shape[1]
        height = array.shape[0]
        print(f"array: {array.shape}")
        if width > 0 and height > 0:
            array = intscale(array, grid_control.zoom_h, grid_control.zoom_w)
                #print("bitmap: %d,%d,3 after scaling: %s" % (height, width, str(array.shape)))
            image = wx.Image(array.shape[1], array.shape[0])
            image.SetData(array.tobytes())
            bmp = wx.Bitmap(image)
            dc.DrawBitmap(bmp, frame_rect.x, frame_rect.y)


class PixelTable(cg.HexTable):
    def __init__(self, control, linked_base):
        self.control = control
        s = self.segment
        self.num_rows = self.converter.calc_grid_height(len(s.data), self.control.bytes_per_row)
        cg.HexTable.__init__(self, s.data, s.style, self.control.items_per_row, self.segment.origin)

    @property
    def segment(self):
        return self.control.segment_viewer.linked_base.segment

    @property
    def converter(self):
        return self.control.pixel_converter

    @property
    def indexes_per_row(self):
        return self.control.bytes_per_row

    def calc_num_rows(self):
        return self.num_rows

    def calc_last_valid_index(self):
        return self.indexes_per_row * self.num_rows

    def prepare_for_drawing(self, start_row, visible_rows, start_cell, visible_cells):
        """Create an array of color index values that contains the visible
        portion of the pixel array

        The array created here is actually bigger than the visible array, as
        wide as the entire set of data, but only as tall as the visible area.
        """
        # cells and columns are equivalent in pixel grids
        s = self.segment
        bytes_per_row = self.control.bytes_per_row
        start_index = start_row * bytes_per_row
        last_index = (start_row + visible_rows) * bytes_per_row
        last_valid = len(s.data)
        if last_index > last_valid:
            total_count = last_index - start_index
            valid_count = last_valid - start_index
            byte_values = np.empty(total_count, dtype=np.uint8)
            byte_values[:valid_count] = s.data[start_index:last_valid]
            byte_values[valid_count:] = 0
            style_values = np.empty(total_count, dtype=np.uint8)
            style_values[:valid_count] = s.style[start_index:last_index]
            style_values[valid_count:] = px.invalid_style
        else:
            byte_values = s.data[start_index:last_index]
            style_values = s.style[start_index:last_index]
        self.current_rectangle = self.converter.calc_color_index_grid(byte_values, style_values, bytes_per_row)

    def rebuild(self):
        segment = self.control.segment_viewer.segment
        self.current_rectangle = None
        self.num_rows = self.converter.calc_grid_height(len(s.data), self.control.bytes_per_row)
        print(f"new num_rows: {self.num_rows}")


class PixelGridControl(SegmentGridControl):
    default_table_cls = PixelTable

    def set_viewer_defaults(self):
        self.bytes_per_row = 16
        self.items_per_row = 8  # 1 byte per pixel fallback
        self.zoom = 2

        self.antic_colors = colors.powerup_colors()
        rgb = colors.calc_playfield_rgb(self.antic_colors)
        highlight_rgb = colors.calc_blended_rgb(rgb, colors.highlight_background_rgb)
        match_rgb = colors.calc_blended_rgb(rgb, colors.match_background_rgb)
        comment_rgb = colors.calc_blended_rgb(rgb, colors.comment_background_rgb)
        data_rgb = colors.calc_dimmed_rgb(rgb, colors.background_rgb, colors.data_background_rgb)
        self.color_list_rgb = (rgb, highlight_rgb, match_rgb, comment_rgb, data_rgb)
        self.empty_color_rgb = colors.empty_background_rgb

    @property
    def pixel_converter(self):
        return self.segment_viewer.pixel_converter

    @property
    def zoom_w(self):
        return self.zoom * self.pixel_converter.scale_width

    @property
    def zoom_h(self):
        return self.zoom * self.pixel_converter.scale_height

    @property
    def pixels_per_byte(self):
        return self.pixel_converter.pixels_per_byte

    @property
    def scale_width(self):
        return self.pixel_converter.scale_width

    @property
    def scale_height(self):
        return self.pixel_converter.scale_height

    def calc_default_table(self, linked_base):
        if hasattr(self, 'segment_viewer'):
            self.items_per_row = self.pixel_converter.validate_pixels_per_row(self.items_per_row)
            self.bytes_per_row = self.pixel_converter.calc_bytes_per_row(self.items_per_row)
            return self.default_table_cls(self, linked_base)
        return SegmentTable(linked_base, self.bytes_per_row)

    def calc_line_renderer(self):
        if hasattr(self, 'segment_viewer'):
            self.items_per_row = self.bytes_per_row * self.pixels_per_byte
            return PixelLineRenderer(self)
        return SegmentGridControl.calc_line_renderer(self)

    def verify_line_renderer(self):
        self.recalc_line_renderer()


class PixelViewer(SegmentViewer):
    name = "pixel"

    ui_name = "Pixel Array"

    control_cls = PixelGridControl

    has_bitmap = True

    has_colors = True

    has_width = True

    width_text = "bitmap width in pixels"

    has_zoom = True

    zoom_text = "bitmap zoom factor"


    def __init__(self, *args, **kwargs):
        SegmentViewer.__init__(self, *args, **kwargs)
        self.pixel_converter = px.AnticE()

    @property
    def window_title(self):
        return "Pixels: " + self.pixel_converter.ui_name

    def update_bitmap(self, evt):
        log.debug("BitmapViewer: machine bitmap changed for %s" % self.control)
        if evt is not Undefined:
            self.control.recalc_view()
            self.linked_base.editor.update_pane_names()

    def validate_width(self, width):
        return width

    def recalc_data_model(self):
        self.control.recalc_view()



import os
import sys

import numpy as np
import wx

from atrip import Segment, style_bits

from sawx.utils.nputil import intscale
from sawx.ui import compactgrid as cg

from ..arch import colors
from ..ui.segment_grid import SegmentGridControl, SegmentVirtualTable
from ..utils import apple2util as a2
from ..editors.linked_base import LinkedBase

from ..viewer import SegmentViewer
from . import bitmap as b
from . import char as c

import logging
log = logging.getLogger(__name__)


class HiresLineRenderer(b.BitmapLineRenderer):
    @classmethod
    def get_image(cls, segment_viewer, bytes_per_row, nr, count, data, style):
        log.debug(f"get_image {bytes_per_row} {nr} {count} {data}")
        subset_count = 14 * len(data)
        bw = np.empty(subset_count, dtype=np.uint8)
        a2.to_560_bw_pixels(data, bw)
        h_colors = colors.get_blended_color_registers([(0, 0, 0), (255, 255, 255)], segment_viewer.preferences.highlight_background_color)
        style_per_pixel = np.repeat(style, 14)

        background = (bw == 0)
        color1 = (bw == 1)
        highlight = (style_per_pixel & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        rgb = np.empty([1, subset_count, 3], dtype=np.uint8)
        rgb[0,background & highlight] = h_colors[0]
        rgb[0,background & np.logical_not(highlight)] = (0, 0, 0)
        rgb[0,color1 & highlight] = h_colors[1]
        rgb[0,color1 & np.logical_not(highlight)] = (255, 255, 255)

        return rgb

    def draw_line(self, grid_control, dc, line_num, col, index, last_index):
        t = grid_control.table
        rect = self.col_to_rect(line_num, col)
        log.debug(f"draw_line: table={t} col={col} indexes:{index}-{last_index}")
        data = t.data[index:last_index]
        style = t.style[index:last_index]
        self.image_cache.draw_item(grid_control, dc, rect, data, style)

    def draw_grid(self, parent, dc, *args, **kwargs):
        try:
            cg.TableLineRenderer.draw_grid(self, parent, dc, *args, **kwargs)
        except IndexError as e:
            log.warning("Index error in draw_grid; likely attempted to render grid on hidden index")


class HiresTable(SegmentVirtualTable):
    row_offset_for_line = a2.hgr_offsets
    segment_name = "Hi-res"
    segment_ui_name = "Apple ][ Hi-res"

    def get_data_style_view(self, linked_base):
        s = linked_base.segment
        byte_order = self.calc_byte_order(s)
        self.apple2_segment = Segment(s, byte_order, name=self.segment_name, verbose_name=self.segment_ui_name)
        data = self.apple2_segment.data
        style = self.apple2_segment.style
        return data, style

    @property
    def segment(self):
        return self.apple2_segment

    def calc_byte_order(self, segment):
        byte_order = a2.hires_byte_order(len(segment))
        return byte_order[0:len(segment)]

    def calc_num_cols(self):
        return 40


class HiresGridControl(b.BitmapGridControl):
    default_table_cls = HiresTable

    def set_viewer_defaults(self):
        self.items_per_row = 40
        self.zoom = 1

    @property
    def bitmap_renderer(self):
        return HiresLineRenderer

    @property
    def pixels_per_byte(self):
        return 7 * 2  # doubling horizontal res for half pixel shifts

    @property
    def scale_width(self):
        return 1

    @property
    def scale_height(self):
        return 2

    def calc_line_renderer(self):
        if hasattr(self, 'segment_viewer'):
            return HiresLineRenderer(self, self.segment_viewer)
        return SegmentGridControl.calc_line_renderer(self)

    ##### Keyboard handling

    def handle_char_move_left(self, evt, flags):
        self.segment_viewer.document.emulator.send_char(0x88)

    def handle_char_move_right(self, evt, flags):
        self.segment_viewer.document.emulator.send_char(0x95)

    def handle_char_move_up(self, evt, flags):
        self.segment_viewer.document.emulator.send_char(0x8b)

    def handle_char_move_down(self, evt, flags):
        self.segment_viewer.document.emulator.send_char(0x8a)


class AppleSegmentChecker:
    @classmethod
    def is_segment_specific_for_display(cls, segment):
        print(f"checking {segment} for origin {cls.segment_origin}: {segment.origin}, size {cls.segment_size}: {len(segment)}")
        return segment.origin == cls.segment_origin and len(segment) == cls.segment_size


class HiresPage1Viewer(AppleSegmentChecker, b.BitmapViewer):
    name = "hgr1"

    ui_name = "Apple ][ Hi-res Page 1"

    viewer_category = "Video"

    control_cls = HiresGridControl

    has_bitmap = True

    has_colors = True

    has_width = False

    has_zoom = True

    zoom_text = "bitmap zoom factor"

    segment_origin = 0x2000

    segment_size = 0x2000

    priority_refresh_frame_count = 1

    @property
    def window_title(self):
        return self.ui_name

    def validate_width(self, width):
        return 560

    def copy_data_from_selections(self):
        segment = self.control.table.segment
        indexes = segment.get_style_indexes(selected=True)
        data = segment[indexes].copy()
        return data

    @property
    def caret_conversion_segment(self):
        return self.control.table.segment


class HiresPage2Viewer(HiresPage1Viewer):
    name = "hgr2"

    ui_name = "Apple ][ Hi-res Page 2"

    segment_origin = 0x4000


class TextTable(HiresTable):
    row_offset_for_line = a2.gr_offsets
    segment_name = "Text"
    segment_ui_name = "Apple ][ Text"

    def calc_byte_order(self, linked_base):
        byte_order = a2.lores_byte_order(len(linked_base.segment))
        return byte_order


class TextGridControl(c.CharGridControl):
    default_table_cls = TextTable

    def set_viewer_defaults(self):
        c.CharGridControl.set_viewer_defaults(self)
        self.items_per_row = 40


class TextPage1Viewer(AppleSegmentChecker, c.CharViewer):
    name = "text1"

    ui_name = "Apple ][ Text Page 1"

    viewer_category = "Video"

    control_cls = TextGridControl

    segment_origin = 0x400

    segment_size = 0x400

    @property
    def window_title(self):
        return self.ui_name


class TextPage2Viewer(TextPage1Viewer):
    name = "text2"

    ui_name = "Apple ][ Text Page 2"

    segment_origin = 0x800



"""Emulator
"""
import os
import sys

import wx

from sawx.action import SawxAction, SawxNameChangeAction, SawxRadioListAction, SawxRadioAction
from sawx.frame import SawxTablessFrame
from sawx.ui.dialogs import prompt_for_dec

from .. import commands
from .. import errors
from ..emulator import find_emulators
from ..documents.emulation_document import EmulationDocument
from ..editors.emulation_editor import EmulationEditor

import logging
log = logging.getLogger(__name__)


class emu_list(SawxRadioListAction):
    def calc_name(self, action_key):
        emu = self.get_item(action_key)
        if emu is None:
            return "Default Based On Disk Image"
        return emu.ui_name

    def calc_list_items(self):
        items = [None]
        items.extend(find_emulators())
        return items

    def calc_checked_list_item(self, action_key, index, item):
        return self.editor.document.emulator_class_override == item

    def perform(self, action_key):
        item = self.get_item(action_key)
        self.editor.document.emulator_class_override = item


class emu_boot_skip_frames(SawxAction):
    def calc_name(self, action_key):
        return "Skip Boot Frames..."

    def show_dialog(self):
        value = prompt_for_dec(self.editor.control, "Skip Boot Frames (decimal, use $ for hex)", "Skip Boot Frames")
        if value is not None:
            print(f"emu_boot_skip_frames: {value}")
            self.editor.document.skip_frames_on_boot = value

    def perform(self, action_key):
        wx.CallAfter(self.show_dialog)


class emu_boot_start_paused(SawxRadioAction):
    name = "Start Paused"
    def calc_checked(self, action_key):
        return self.editor.document.pause_emulator_on_boot

    def perform(self, action_key):
        self.editor.document.pause_emulator_on_boot = not self.editor.document.pause_emulator_on_boot
        print("NOW Paused?", self.editor.document.pause_emulator_on_boot)


class emu_boot_disk_image(SawxAction):
    def calc_name(self, action_key):
        return "Boot Disk Image"

    def do_boot(self, doc, source_document):
        doc.boot(source_document.collection.containers[0])

    def perform(self, action_key):
        print(f"Booting disk image using {self.editor.document.emulator_class_override}")
        source_document = self.editor.document
        try:
            doc = EmulationDocument.create_document(source_document, source_document.emulator_class_override)
        except errors.EmulatorInUseError as e:
            print(f"Currently running emulator: {e.current_emulator_document}")
            doc = e.current_emulator_document
            doc.pause_emulator()
            if self.editor.frame.confirm(f"Only one {doc.emulator.ui_name} emulator can be running at any one time.\n\nReplace with the new boot image? Current emulation\nwill be lost.", "Replace Emulator"):
                self.do_boot(doc, source_document)
                editor = wx.GetApp().find_editor_of_document(doc)
                editor.frame.Raise()
                doc.recalc_event(True)
            else:
                doc.resume_emulator()
        else:
            self.do_boot(doc, source_document)
            print(f"emulator document: {doc}")
            editor = EmulationEditor(doc)
            frame = SawxTablessFrame(editor)
            frame.Show()

class emu_boot_segment(emu_boot_disk_image):
    def calc_name(self, action_key):
        return "Boot Segment"

    def do_boot(self, doc, source_document):
        doc.boot(self.editor.focused_viewer.segment)


class emu_restore(SawxAction):
    def calc_name(self, action_key):
        return "Restore Saved State..."

    def perform(self, action_key):
        print(f"Restoring saved state from file")


class emu_pause_resume(SawxNameChangeAction):
    """Stop or restart the emulation
    """
    def calc_name(self, action_key):
        if not self.editor.document.emulator_running:
            name = "Resume"
        else:
            name = "Pause"
        return name

    def perform(self, action_key):
        if self.editor.document.emulator_running:
            self.editor.document.pause_emulator()
        else:
            self.editor.document.resume_emulator()


class emu_step(SawxAction):
    """Step to next instruction, into subroutine
    """
    def calc_name(self, action_key):
        return "Step"

    def calc_enabled(self, action_key):
        return self.editor.document.emulator.emulator_paused

    def perform(self, action_key):
        self.editor.document.debugger_step()


class emu_step_over(emu_step):
    """Step to next line, through subroutines
    """
    def calc_name(self, action_key):
        return "Step Into"

    def perform(self, action_key):
        print("resume!")


class emu_step_out(emu_step):
    """Step to subroutine return statement, through any subroutines
    called by the current block of code.
    """
    def calc_name(self, action_key):
        return "Step Over"

    def perform(self, action_key):
        print("resume!")


class emu_break_next_scan_line(SawxAction):
    """Continue to next scan line
    """
    def calc_name(self, action_key):
        return "Break at Next Scan Line"

    def calc_enabled(self, action_key):
        return self.editor.document.emulator.emulator_paused

    def perform(self, action_key):
        self.editor.document.debugger_break_next_scan_line()


class emu_break_vbi_start(SawxAction):
    """Continue and break at the start of the next VBI
    """
    def calc_name(self, action_key):
        return "Break at Next VBI Start"

    def calc_enabled(self, action_key):
        return self.editor.document.emulator.emulator_paused

    def perform(self, action_key):
        self.editor.document.debugger_break_vbi_start()


class emu_break_vbi_end(SawxAction):
    """Continue and break at the end of the next VBI
    """
    def calc_name(self, action_key):
        return "Break at Next VBI End"

    def calc_enabled(self, action_key):
        return self.editor.document.emulator.emulator_paused

    def perform(self, action_key):
        self.editor.document.debugger_break_vbi_end()


class emu_break_dli_start(SawxAction):
    """Continue and break at the start of the next DLI
    """
    def calc_name(self, action_key):
        return "Break at Next DLI Start"

    def calc_enabled(self, action_key):
        return self.editor.document.emulator.emulator_paused

    def perform(self, action_key):
        self.editor.document.debugger_break_dli_start()


class emu_break_dli_end(SawxAction):
    """Continue and break at the end of the next DLI
    """
    def calc_name(self, action_key):
        return "Break at Next DLI End"

    def calc_enabled(self, action_key):
        return self.editor.document.emulator.emulator_paused

    def perform(self, action_key):
        self.editor.document.debugger_break_dli_end()


class emu_break_frame(emu_step):
    """Continue and break when reaching the end of the current frame
    """
    def calc_name(self, action_key):
        return "Break at End of Frame"

    def perform(self, action_key):
        self.editor.document.debugger_count_frames()


class emu_a8_start_key(SawxAction):
    """Press the Start button
    """
    def calc_name(self, action_key):
        return "Start"

    def perform(self, action_key):
        print("START!")
        self.editor.document.emulator.forced_modifier = "start"


class emu_a8_select_key(SawxAction):
    """Press the Select button
    """
    def calc_name(self, action_key):
        return "Select"

    def perform(self, action_key):
        self.editor.document.emulator.forced_modifier = "select"


class emu_a8_option_key(SawxAction):
    """Press the Option button
    """
    def calc_name(self, action_key):
        return "Option"

    def perform(self, action_key):
        self.editor.document.emulator.forced_modifier = "option"


class emu_coldstart(SawxAction):
    """Simulate turning off the power and turning it back on again
    """
    def calc_name(self, action_key):
        return "Reboot"

    def perform(self, action_key):
        self.editor.document.emulator.coldstart()
        self.editor.document.resume_emulator()


class emu_warmstart(SawxAction):
    """Simulate pressing the system reset key
    """
    def calc_name(self, action_key):
        return "System Reset"

    def perform(self, action_key):
        self.editor.document.emulator.warmstart()


class emu_prev_history(SawxAction):
    """Load the previous saved frame into the current state of the emulator
    """
    def calc_name(self, action_key):
        return "Previous Save Point"

    def perform(self, action_key):
        d = self.editor.document
        if d.emulator_running:
            d.pause_emulator()
        else:
            d.history_previous()


class emu_next_history(SawxAction):
    """Load the next saved frame into the current state of the emulator
    """
    def calc_name(self, action_key):
        return "Next Save Point"

    def perform(self, action_key):
        d = self.editor.document 
        print("OEUOEU", d.emulator_running)
        if d.emulator_running:
            d.pause_emulator()
        else:
            print("BBVJKBVJKBQJBKQB")
            d.history_next()



#!/usr/bin/env python
import os
import glob
import json
import hashlib
import uuid

from PIL import Image
import numpy as np

import logging
log = logging.getLogger(__name__)


def bits_from_image(image):
    raw = image.tobytes()
    array = np.array(image)
    # print(array.shape)
    # print(array[:,16:32])

    line = np.empty((8, 1024), dtype=np.uint8)
    line[:,0:256] = array[0:8,0:256]
    line[:,256:512] = array[8:16,0:256]
    line[:,512:768] = array[16:24,0:256]
    line[:,768:1024] = array[24:32,0:256]

    # print(line.shape)
    # for i in range(32):
    #     print(line[:,8*i:8*i + 8])

    byte_array = np.packbits(line).reshape((8,128))
    # print(byte_array.shape)
    # for i in range(32):
    #     print(byte_array[:,i])
    return np.transpose(byte_array).copy()


if __name__ == "__main__":
    import sys

    source_dir = "../../Eightbit-Atari-Fonts/Original Bitmaps"
    template_dir = "../omnivore/templates"

    if len(sys.argv) > 1:
        # print out data for a particular font
        for font in sys.argv[1:]:
            filename = os.path.join(source_dir, font + ".png")
            im = Image.open(filename)
            data = bits_from_image(im)
            for i in range(128):
                row = data[i]
                print("        .byte " + ",".join([f"${a:02x}" for a in row]))
    else:
        for filename in sorted(glob.glob(source_dir + "/*")):
            name, _ = os.path.splitext(os.path.basename(filename))

            im = Image.open(filename)
            # print(im)

            data_bytes = bits_from_image(im)
            md5 = hashlib.md5(data_bytes).digest()
            new_uuid = str(uuid.UUID(bytes=md5))
            # print("md5", new_uuid)
            inf = {
                'uuid': new_uuid,
                'name': f"8x8 Atari 8-bit Font: {name}",
                'char_w': 8,
                'char_h': 8,
                'font_group': "Steve Boswell's Eightbit-Atari-Fonts",
            }

            basename = os.path.join(template_dir, str(new_uuid) + ".font")
            print(f"{name}: writing {basename}")
            with open(basename, 'wb') as fh:
                fh.write(data_bytes)
            with open(basename + ".inf", 'w') as fh:
                text = json.dumps(inf, indent=4, sort_keys=True) + "\n"
                fh.write(text)



import numpy as np

from . import errors
from .ataridos import AtariDosDirent, AtariDosDiskImage, XexSegment
from .segments import DefaultSegment, EmptySegment, ObjSegment, RawSectorsSegment, SegmentSaver

import logging
log = logging.getLogger(__name__)


class SpartaDosDirent(AtariDosDirent):
    format = np.dtype([
        ('status', 'u1'),
        ('sector', '<u2'),
        ('len_l', '<u2'),
        ('len_h', 'i1'),
        ('filename','S8'),
        ('ext','S3'),
        ('date','S3'),
        ('time','S3'),
        ])

    def __init__(self, image, file_num=0, bytes=None, starting_sector=None):
        self.length = 0
        self.sector_map = None
        self.sector_map_index = 0
        AtariDosDirent.__init__(self, image, file_num, bytes=bytes)
        if starting_sector is not None:
            # Root directory doesn't have the starting sector in the dirent,
            # rather the boot sector so it must be specified here.
            self.starting_sector = starting_sector
            self.is_sane = self.sanity_check(image)

    def __str__(self):
        output = "o" if self.opened_output else "."
        subdir = "D" if self.is_dir else "."
        in_use = "u" if self.in_use else "."
        deleted = "d" if self.deleted else "."
        locked = "*" if self.locked else " "
        flags = "%s%s%s%s%s %03d" % (output, subdir, in_use, deleted, locked, self.starting_sector)
        return "File #%-2d (%s) %-8s%-3s  %8d  %s" % (self.file_num, flags, self.basename.decode('latin1'), self.ext.decode('latin1'), self.length, self.str_timestamp)

    @property
    def verbose_info(self):
        flags = []
        if self.opened_output: flags.append("OUT")
        if self.is_dir: flags.append("DIR")
        if self.in_use: flags.append("IN_USE")
        if self.deleted: flags.append("DEL")
        if self.locked: flags.append("LOCK")
        return "flags=[%s]" % ", ".join(flags)

    def parse_raw_dirent(self, image, data):
        if data is None:
            return
        values = data.view(dtype=self.format)[0]
        flag = values[0]
        self.flag = flag
        self.locked = (flag&0x1) > 0
        self.hidden = (flag&0x10) > 0
        self.archived = (flag&0x100) > 0
        self.in_use = (flag&0b1000) > 0
        self.deleted = (flag&0b10000) > 0
        self.is_dir = (flag&0b100000) > 0
        self.opened_output = (flag&0b10000000) > 0
        self.starting_sector = int(values[1])
        self.basename = bytes(values[4]).rstrip()
        if self.is_dir:
            self.ext = b""
        else:
            self.ext = bytes(values[5]).rstrip()
        self.length = 256*256*values[3] + values[2]
        self.date_array = tuple(data[17:20])
        self.time_array = tuple(data[20:23])
        self.is_sane = self.sanity_check(image)

    def sanity_check(self, image):
        if not self.in_use:
            return True
        if not image.header.sector_is_valid(self.starting_sector):
            return False
        return True

    @property
    def str_timestamp(self):
        str_date = "%d/%d/%d" % self.date_array
        str_time = "%d:%d:%d" % self.time_array
        return "%s %s" % (str_date, str_time)

    def start_read(self, image):
        if not self.is_sane:
            log.debug("Invalid directory entry '%s', starting_sector=%s" % (str(self), self.starting_sector))
            raise errors.InvalidDirent("Invalid directory entry '%s'" % str(self))
        self.sector_map = image.get_sector_map(self.starting_sector)
        self.sector_map_index = 0
        self.length_remaining = self.length

    def read_sector(self, image):
        sector = self.sector_map[self.sector_map_index]
        if sector == 0:
            return None, True, 0, self.length_remaining
        raw, pos, size = image.get_raw_bytes(sector)
        num_data_bytes = min(self.length_remaining, size)
        self.length_remaining -= num_data_bytes
        self.sector_map_index += 1
        return raw[0:num_data_bytes], sector == 0, pos, num_data_bytes


class SpartaDosDiskImage(AtariDosDiskImage):
    def __init__(self, *args, **kwargs):
        self.first_bitmap = 0
        self.num_bitmap = 0
        self.root_dir = 0
        self.root_dir_dirent = None
        self.fs_version = 0
        AtariDosDiskImage.__init__(self, *args, **kwargs)

    def __str__(self):
        return "%s Sparta DOS Format: %d usable sectors (%d free), %d files" % (self.header, self.total_sectors, self.unused_sectors, len(self.files))

    boot_record_type = np.dtype([
        ('unused', 'u1'),
        ('num_boot', 'u1'),
        ('boot_addr', '<u2'),
        ('init_addr', '<u2'),
        ('jmp', 'u1'),
        ('cont_addr', '<u2'),
        ('root_dir','<u2'),
        ('num_sectors','<u2'),
        ('num_free','<u2'),
        ('num_bitmap', 'u1'),
        ('bitmap','<u2'),
        ('first_free', '<u2'),
        ('first_free_dir','<u2'),
        ('vol_name','S8'),
        ('num_tracks','u1'),
        ('sector_size','u1'),
        ('fs_version','u1'),
        ])

    sector_size_map = {0: 256,
                       1: 512,
                       0x80: 128,
                       }

    def get_boot_sector_info(self):
        data, style = self.get_sectors(1)
        values = data[0:33].view(dtype=self.boot_record_type)[0]
        self.num_boot = values['num_boot']
        self.boot_addr = values['boot_addr']
        self.first_bitmap = values['bitmap']
        self.num_bitmap = values['num_bitmap']
        self.root_dir = values['root_dir']
        self.fs_version = values['fs_version']
        s = values['sector_size']
        self.sector_size = self.sector_size_map.get(values['sector_size'], -1)
        self.total_sectors = values['num_sectors']
        self.unused_sectors = values['num_free']
        num = self.header.max_sectors
        self.is_sane = self.total_sectors == num and values['first_free'] <= num and self.first_bitmap <= num and self.root_dir <= num and self.fs_version in [0x11, 0x20, 0x21] and self.sector_size != -1
        if not self.is_sane:
            raise errors.InvalidDiskImage("Invalid SpartaDos parameters in boot header")

    def get_vtoc(self):
        pass

    def get_directory(self):
        self.files = []
        dir_map = self.get_sector_map(self.root_dir)
        sector = dir_map[0]
        if sector == 0:
            return
        bytes, pos, size = self.get_raw_bytes(sector)
        d = SpartaDosDirent(self, 0, bytes[0:23], starting_sector=self.root_dir)
        s = self.get_file_segment(d)
        for filenum, i in enumerate(range(23, d.length, 23)):
            dirent = SpartaDosDirent(self, filenum + 1, s[i:i + 23])
            self.files.append(dirent)
        self.root_dir_dirent = d

    def get_boot_segments(self):
        segments = []
        num = min(self.num_boot, 1)
        s = self.get_sector_slice(1, num)
        r = self.rawdata[s]
        addr = self.boot_addr
        header = ObjSegment(r[0:43], 0, 0, addr, addr + 43, name="Boot Header")
        segments.append(header)
        if self.num_boot > 0:
            sectors = ObjSegment(r, 0, 0, addr, addr + len(r), name="Boot Sectors")
            code = ObjSegment(r[43:], 0, 0, addr + 43, addr + len(r), name="Boot Code")
            segments.extend([header, code])
        return segments

    def get_vtoc_segments(self):
        r = self.rawdata
        segments = []
        addr = 0
        start, count = self.get_contiguous_sectors(self.first_bitmap, self.num_bitmap)
        if self.sector_size == 512:
            num_boot = 1
            boot_size = 512
        else:
            num_boot = 3
            boot_size = 128
        segment = RawSectorsSegment(r[start:start+count], self.first_bitmap, self.num_bitmap, count, 0, 0, self.sector_size, name="Bitmap")
        segments.append(segment)
        return segments

    def get_sector_map(self, sector):
        m = None
        while sector > 0:
            b, _ = self.get_sectors(sector)
            sector, prev = b[0:4].view(dtype='<u2')
            if m is None:
                m = np.copy(b[4:].view(dtype='<u2'))
            else:
                m = np.hstack((m, b[4:].view(dtype='<u2')))
        return m

    def get_directory_segments(self):
        dirent = self.root_dir_dirent
        segment = self.get_file_segment(dirent)
        segment.name = dirent.filename
        segment.map_width = 23
        segments = [segment]
        return segments

    def get_file_segment(self, dirent):
        byte_order = []
        dirent.start_read(self)
        while True:
            bytes, last, pos, size = dirent.read_sector(self)
            if not last:
                byte_order.extend(list(range(pos, pos + size)))
            else:
                break
        if len(byte_order) > 0:
            name = "%s %d@%d %s" % (dirent.filename, dirent.length, dirent.starting_sector, dirent.str_timestamp)
            verbose_name = "%s (%d bytes, sector map@%d) %s %s" % (dirent.filename, dirent.length, dirent.starting_sector, dirent.verbose_info, dirent.str_timestamp)
            raw = self.rawdata.get_indexed(byte_order)
            segment = DefaultSegment(raw, name=name, verbose_name=verbose_name)
        else:
            segment = EmptySegment(self.rawdata, name=dirent.filename, error=dirent.str_timestamp)
        return segment



import os
import glob

import numpy as np
import pytest
import jsonpickle

from mock import globbed_sample_atari_files, globbed_sample_atari_collections

from atrip.collection import Collection
from atrip.container import Container
from atrip.segment import Segment
import atrip.errors as errors


class TestCollection:
    def test_create(self):
        data = np.arange(4096, dtype=np.uint8)
        container = Container(data)
        collection = Collection("test", container=container)
        print(collection)
        assert collection.archiver.__class__.__name__ == "PlainFileArchiver"
        assert len(collection.containers) == 1
        assert len(collection.containers[0]) == 4096

    def test_serialize(self):
        filename = "dos_sd_test1.atr"
        pathname = os.path.join(os.path.dirname(__file__), "../samples", filename)
        c = Collection(pathname)
        m = c.containers[0].media
        m.set_comment_at(100, "at location 100")
        m.set_comment_at(1000, "at location 1000")
        m.set_comment_at(10000, "at location 10000")
        s = {}
        c.serialize_session(s)
        print(s)

        j = jsonpickle.dumps(s)
        print(j)
        
        sprime = jsonpickle.loads(j)
        jprime = jsonpickle.dumps(sprime)
        print(jprime)

        assert j == jprime

        c2 = Collection(pathname, session=sprime)
        s2 = {}
        m2 = c2.containers[0].media
        assert m2.get_comment_at(100) == "at location 100"
        assert m2.get_comment_at(1000) == "at location 1000"
        assert m2.get_comment_at(10000) == "at location 10000"
        c2.serialize_session(s2)
        j2 = jsonpickle.dumps(s2)
        assert j == j2

    @pytest.mark.parametrize(("pathname"), globbed_sample_atari_files)
    def test_single(self, pathname):
        collection = Collection(pathname)
        print(collection.verbose_info)
        output = "tmp." + os.path.basename(pathname)
        output = os.path.join(os.path.dirname(__file__), output)
        try:
            collection.save(output)
        except errors.InvalidAlgorithm as e:
            pytest.skip(f"skipping {pathname}: {e}")
        else:
            # compressed data may not be the same; don't really care as long as
            # it decompresses the same
            collection2 = Collection(output)
            print(collection.decompression_order)
            print(collection2.decompression_order)
            assert collection.decompression_order == collection2.decompression_order
            for c, c2 in zip(collection.containers, collection2.containers):
                print(f"checking collections: {c}, {c2}")
                assert np.array_equal(c._data, c2._data)

            if ".atr." in pathname or ".dcm." in pathname:
                if ".atr." in pathname:
                    ext = ".atr"
                else:
                    ext = ".dcm"
                orig, _ = pathname.split(ext + ".", 1)
                orig += ext
                collection_uncompressed = Collection(orig)
                print("comparing against uncompressed\n", collection_uncompressed.verbose_info)
                container_orig = collection.containers[0]
                container_uncompressed = collection_uncompressed.containers[0]
                assert np.array_equal(container_orig._data, container_uncompressed._data)

            if ".zip." in pathname or ".tar." in pathname:
                if ".zip." in pathname:
                    ext = ".zip"
                else:
                    ext = ".tar"
                orig, _ = pathname.split(ext + ".", 1)
                orig += ext
                collection_uncompressed = Collection(orig)
                print("comparing against uncompressed\n", collection_uncompressed.verbose_info)
                assert len(collection.containers) == len(collection_uncompressed.containers)
                for c, c2 in zip(collection.containers, collection_uncompressed.containers):
                    assert np.array_equal(c._data, c2._data)

    @pytest.mark.parametrize(("pathname"), globbed_sample_atari_collections)
    def test_multiple(self, pathname):
        self.test_single(pathname)

if __name__ == "__main__":
    t = TestCollection()
    # t.test_serialize()
    # t.test_single("../samples/mydos_sd_mydos4534.dcm.lz4")
    # t.test_single("../samples/dos_sd_test1.atr.gz")
    # t.test_single("../samples/dos_sd_test_collection.zip")
    # t.test_single("../samples/dos_sd_test_collection.tar")
    # t.test_single("../samples/dos_sd_test_collection.zip.lz4")
    t.test_create()



import numpy as np

from .. import errors
from ..file_type import FileType
from ..segment import Segment
from ..style_bits import get_style_bits

import logging
log = logging.getLogger(__name__)


class Dos33Bsave(FileType):
    """Parse a binary chunk into segments according to the DOS 3.3 binary
    dump format
    """
    ui_name = "Apple DOS 3.3 Object File"

    def calc_segments(self):
        """Convenience method used by subclasses to create any sub-segments
        within this segment.

        """
        origin = self[0] + 256 * self[1]
        expected_count = self[2] + 256 * self[3]
        count = len(self) - 4
        if count != expected_count:
            raise errors.InvalidBinaryFile(f"Extra data after BSAVE segment: found {count}, header specifies {expected_count} bytes")
        s = Segment(self, 4, name=f"BSAVE data: ${count:04x}@${origin:04x}", origin=origin, length=count)
        return [s]



import sys

import wx

from sawx.preferences import SawxEditorPreferences
from sawx.ui.fonts import str_to_font, default_font


def byte2hex(val):
    return "$02x" % val

def word2hex(val):
    return "$04x" % val

def int2hex(val):
    return "%x" % val

byte2str = byte2hex
word2str = word2hex
int2str = int2hex

class ByteEditorPreferences(SawxEditorPreferences):
    def set_defaults(self):
        SawxEditorPreferences.set_defaults(self)

        # int_display_format = Enum(
        #     "Hexadecimal",
        #     "Decimal",
        #     "Both",
        # )

        # hex_display_format = Enum(
        #     "$XX",
        #     "0xXX",
        #     "XXh",
        # )

        self.hex_grid_lower_case = True
        self.assembly_lower_case = False
        self.disassembly_column_widths = (0, 0, 0)
        self.default_cpu = "6502"
        self.map_width_low = 1
        self.map_width_high = 256
        self.map_width = 40
        self.bitmap_width_low = 1
        self.bitmap_width_high = 16
        self.bitmap_width = 1
        self.hex_grid_width_low = 1
        self.hex_grid_width_high = 256
        self.hex_grid_width = 16
        self.text_copy_stringifier = "basic_data"

    @property
    def hex_format_character(self):
        return "x" if self.hex_grid_lower_case else "X"



from .collection import Collection
from . import errors

import logging
log = logging.getLogger(__name__)


def identify_loader(file_guess):
    mime_type = None
    try:
        data = file_guess.all_data
    except IOError as e:
        log.debug(f"atrip loader: error reading entire file: {e}")
    else:
        try:
            collection = Collection(file_guess.uri, data)
        except (errors.UnsupportedDiskImage, IOError) as e:
            log.debug(f"error in atrip parser: {e}")
        else:
            log.debug(f"{collection}")

        if collection.mime_type:
            log.debug(f"atrip loader: identified {collection.mime_type}")
            return dict(mime=collection.mime_type, ext="", atrip_collection=collection)
        else:
            log.debug(f"atrip loader: not recognized")
    return None



import numpy as np

from . import SetSelectionCommand, SetRangeCommand, ChangeStyleCommand, SetRangeValueModifyIndexesCommand
from sawx.utils.permute import bit_reverse_table

from atrip.disassembler import mini_assemble

import logging
log = logging.getLogger(__name__)


class SetValueCommand(SetSelectionCommand):
    short_name = "set_value"
    ui_name = "Set Value"


class ZeroCommand(SetSelectionCommand):
    short_name = "zero"
    ui_name = "Zero Bytes"

    def __init__(self, segment, current_selection):
        super().__init__(segment, current_selection, 0)


class FFCommand(SetSelectionCommand):
    short_name = "ff"
    ui_name = "FF Bytes"

    def __init__(self, segment, current_selection):
        super().__init__(segment, current_selection, 0xff)


class NOPCommand(SetSelectionCommand):
    short_name = "nop"
    ui_name = "NOP Bytes"


class SetHighBitCommand(SetSelectionCommand):
    short_name = "set_high_bit"
    ui_name = "Set High Bit"

    def get_data(self, orig):
        return np.bitwise_or(orig, 0x80)


class ClearHighBitCommand(SetSelectionCommand):
    short_name = "clear_high_bit"
    ui_name = "Clear High Bit"

    def get_data(self, orig):
        return np.bitwise_and(orig, 0x7f)


class BitwiseNotCommand(SetSelectionCommand):
    short_name = "bitwise_not"
    ui_name = "Bitwise NOT"

    def get_data(self, orig):
        return np.invert(orig)


class OrWithCommand(SetSelectionCommand):
    short_name = "or_value"
    ui_name = "OR With"

    def get_data(self, orig):
        return np.bitwise_or(orig, self.data)


class AndWithCommand(SetSelectionCommand):
    short_name = "and_value"
    ui_name = "AND With"

    def get_data(self, orig):
        return np.bitwise_and(orig, self.data)


class XorWithCommand(SetSelectionCommand):
    short_name = "xor_value"
    ui_name = "XOR With"

    def get_data(self, orig):
        return np.bitwise_xor(orig, self.data)


class LeftShiftCommand(SetSelectionCommand):
    short_name = "left_shift"
    ui_name = "Left Shift"

    def get_data(self, orig):
        return np.left_shift(orig, 1)


class RightShiftCommand(SetSelectionCommand):
    short_name = "right_shift"
    ui_name = "Right Shift"

    def get_data(self, orig):
        return np.right_shift(orig, 1)


class LeftRotateCommand(SetSelectionCommand):
    short_name = "left_rotate"
    ui_name = "Left Rotate"

    def get_data(self, orig):
        rotated = np.right_shift(np.bitwise_and(orig, 0x80), 7)
        return np.bitwise_or(np.left_shift(orig, 1), rotated)


class RightRotateCommand(SetSelectionCommand):
    short_name = "right_rotate"
    ui_name = "Right Rotate"

    def get_data(self, orig):
        rotated = np.left_shift(np.bitwise_and(orig, 0x01), 7)
        return np.bitwise_or(np.right_shift(orig, 1), rotated)


class ReverseBitsCommand(SetSelectionCommand):
    short_name = "reverse_bits"
    ui_name = "Reverse Bits"

    def get_data(self, orig):
        return bit_reverse_table[orig]


class RandomBytesCommand(SetSelectionCommand):
    short_name = "random_bytes"
    ui_name = "Random Bytes"

    def get_data(self, orig):
        return np.random.randint(0, 256, len(orig), dtype=np.uint8)


class RampUpCommand(SetSelectionCommand):
    short_name = "ramp_up"
    ui_name = "Ramp Up"

    def get_data(self, orig):
        num = np.alen(orig)
        if type(self.data) == slice:
            first = self.data.start
            last = self.data.stop
            if last == -1:
                last = self.data.start + num
        else:
            first = self.data
            last = self.data + num
        step = (last - first) / num
        print(f"range: {first}, {last}, {step}")
        return np.arange(first, last, step)


class RampDownCommand(SetSelectionCommand):
    short_name = "ramp_down"
    ui_name = "Ramp Down"

    def get_data(self, orig):
        num = np.alen(orig)
        if type(self.data) == slice:
            first = self.data.start
            last = self.data.stop
            if last == -1:
                last = self.data.start - num
        else:
            first = self.data
            last = self.data - num
        step = (last - first) / num
        return np.arange(first, last, step)


class AddValueCommand(SetSelectionCommand):
    short_name = "add_value"
    ui_name = "Add"

    def get_data(self, orig):
        return orig + self.data


class SubtractValueCommand(SetSelectionCommand):
    short_name = "subtract_value"
    ui_name = "Subtract"

    def get_data(self, orig):
        return orig - self.data


class SubtractFromCommand(SetSelectionCommand):
    short_name = "subtract_from"
    ui_name = "Subtract From"

    def get_data(self, orig):
        return self.data - orig


class MultiplyCommand(SetSelectionCommand):
    short_name = "multiply"
    ui_name = "Multiply"

    def get_data(self, orig):
        return orig * self.data


class DivideByCommand(SetSelectionCommand):
    short_name = "divide"
    ui_name = "Divide By"

    def get_data(self, orig):
        return orig // self.data


class DivideFromCommand(SetSelectionCommand):
    short_name = "divide_from"
    ui_name = "Divide From"

    def get_data(self, orig):
        return self.data // orig


class ReverseSelectionCommand(SetSelectionCommand):
    short_name = "reverse_selection"
    ui_name = "Reverse Selection"

    def get_data(self, orig):
        return orig[::-1,...]


class ReverseGroupCommand(SetSelectionCommand):
    short_name = "reverse_group"
    ui_name = "Reverse In Groups"

    def get_data(self, orig):
        num = len(orig)
        chunk = self.data
        num_groups = num // chunk
        indexes = np.arange(num)
        if num_groups > 0:
            # have to handle special case: can't do indexes[9:-1:-1] !!!
            indexes[0:chunk] = indexes[self.data-1::-1]
            for i in range(1,num_groups):
                start = i * chunk
                indexes[start:start+chunk] = indexes[start+chunk-1:start-1:-1]
        print(num, chunk, num_groups, indexes)
        return orig[indexes]


class ApplyTraceSegmentCommand(ChangeStyleCommand):
    short_name = "applytrace"
    ui_name = "Apply Trace to Segment"

    def get_style(self, editor):
        v = editor.focused_viewer
        trace, mask = v.get_trace(save=True)
        self.clip(trace)
        style_data = (self.segment.style[self.start_index:self.end_index].copy() & mask) | trace
        return style_data

    def set_undo_flags(self, flags):
        flags.byte_values_changed = True
        flags.index_range = self.start_index, self.end_index


class ClearTraceCommand(ChangeStyleCommand):
    short_name = "cleartrace"
    ui_name = "Clear Current Trace Results"

    def get_style(self, editor):
        mask = self.segment.get_style_mask(match=True)
        style_data = (self.segment.style[:].copy() & mask)
        return style_data

    def update_can_trace(self, editor):
        editor.can_trace = False


class SetLabelCommand(ChangeStyleCommand):
    short_name = "set_comment"
    ui_name = "Label"
    serialize_order =  [
            ('segment', 'int'),
            ('addr', 'addr'),
            ('label', 'string'),
            ]

    def __init__(self, segment, addr, label):
        ChangeStyleCommand.__init__(self, segment)
        self.addr = addr
        self.label = label

    def __str__(self):
        if len(self.label) > 20:
            text = self.label[:20] + "..."
        else:
            text = self.label
        return "%s: %s" % (self.ui_name, text)

    def do_change(self, editor, undo):
        old = self.segment.memory_map.get(self.addr, None)
        self.segment.memory_map[self.addr] = self.label
        return old

    def undo_change(self, editor, old_data):
        if old_data is None:
            self.segment.memory_map.pop(self.addr, "")
        else:
            self.segment.memory_map[self.addr] = old_data


class ClearLabelCommand(ChangeStyleCommand):
    short_name = "clear_comment"
    ui_name = "Remove Label"
    serialize_order =  [
            ('segment', 'int'),
            ('ranges', 'int_list'),
            ]

    def __init__(self, segment, ranges):
        ChangeStyleCommand.__init__(self, segment)
        print(ranges)
        self.ranges = ranges

    def do_change(self, editor, undo):
        print(self.ranges)
        indexes = ranges_to_indexes(self.ranges)
        origin = self.segment.origin
        old = {}
        for i in indexes:
            addr = i + origin
            old[addr] = self.segment.memory_map.get(addr, None)
            self.segment.memory_map.pop(addr, "")
        return old

    def undo_change(self, editor, old_data):
        if old_data is not None:
            indexes = ranges_to_indexes(self.ranges)
            for addr, label in old_data.iteritems():
                if label is None:
                    self.segment.memory_map.pop(addr, "")
                else:
                    self.segment.memory_map[addr] = label
